from __future__ import annotations
import uvicore
import inspect
from copy import copy
from functools import partial
from uvicore.contracts import Router as RouterInterface
from uvicore.contracts import Package as PackageInterface
from uvicore.contracts import Routes as RoutesInterface
from uvicore.support.module import load
from uvicore.support import str as string
from uvicore.support.dumper import dump, dd
from uvicore.http.routing.guard import Guard
from uvicore.support.dictionary import deep_merge
from uvicore.typing import Dict, Callable, List, TypeVar, Generic, Decorator, Optional, Union

# Generic Route (Web or Api)
R = TypeVar('R')


@uvicore.service()
class Router(Generic[R], RouterInterface[R]):
    """Abstract base router class for Web and Api Router Implimentations"""

    @property
    def package(self) -> PackageInterface:
        return self._package

    @property
    def routes(self) -> Dict[str, R]:
        return self._routes

    def __init__(self, package: PackageInterface, prefix: str, name: str = '', controllers: str = None):
        # Instance variables
        self._package = package
        self._routes = Dict()
        self._tmp_routes = Dict()
        self.controllers: str = controllers

        # Prefix
        if prefix == '/': prefix = ''
        if prefix:
            if prefix[-1] == '/': prefix = prefix[0:-1]
            if prefix[0] != '/': prefix = '/' + prefix
        self.prefix = prefix

        # Clean name
        if name:
            name = name.replace('/', '.')
            if name[-1] == '.': name = name[0:-1]  # Remove trailing .
            if name[0] == '.': name = name[1:]     # Remove beginning .
        self.name = name

    def controller(self,
        module: Union[str, Callable],
        *,
        prefix: str = '',
        name: str = '',
        tags: Optional[List[str]] = None,
        options: Dict = {}
    ) -> List:
        """Include a Route Controller"""
        if prefix:
            if prefix[-1] == '/': prefix = prefix[0:-1]  # Remove trailing /
            if prefix[0] != '/': prefix = '/' + prefix   # Add beginning /

        # Get name
        if not name: name = prefix

        # Clean name
        if name:
            name = name.replace('/', '.')
            if name[-1] == '.': name = name[0:-1]  # Remove trailing .
            if name[0] == '.': name = name[1:]     # Remove beginning .

        # Import controller module from string
        cls = module
        if type(module) == str:
            if self.controllers:
                if '.' not in module:
                    # We are defining just 'home', so we add .Home class
                    module = self.controllers + '.' + module + '.' + string.studly(module)
                elif module[0] == '.':
                    # We are appending the path to self.controllers
                    # Must have class too, ex: .folder.stuff.Stuff
                    module = self.controllers + module
                # elif module.count('.') == 1: # NO, we'll just add a . before, like .home.Home and it does the same thing
                #     # We are defining the file and the class (home.Home)
                #     # Only works with ONE dot.  If you want to append use .folder.stuff.Stuff
                #     module = self.controllers + '.' + module
                else:
                    # We are defining the FULL module path even though we have defined a self.controller path
                    # Must have class too, ex: acme.appstub.http.api.stuff.Stuff
                    pass

            # Dynamically import the calculated module
            cls = load(module).object
            if str(type(cls)) == "<class 'module'>":
                # Trying to load a module, but we want the class inside the module, auto add
                module = module + '.' + string.studly(module.split('.')[-1])
                cls = load(module).object

        # Instantiate controller file
        controller: Routes = cls(self.package, **options)

        # New self (Web or Api) router instance
        router = self.__class__(self.package, self.prefix + prefix, self.name + '.' + name, self.controllers)

        # Register controllers routes and return new updated router
        router = controller.register(router)

        # Add contoller class level attributes as middleware to each route on this controller
        controller_middlewares = controller._middleware()
        if controller_middlewares:
            for route in router.routes.values():
                (route.middleware, route.endpoint) = self._merge_route_middleware(controller_middlewares, route.middleware, route.endpoint)

        # Merge controllers routes into this main (parent of recursion) router
        #dump(router.routes)
        self.routes.merge(router.routes)

        # Return just this controllers routes as a list
        routes = []
        for route in router.routes.keys():
            # Append .controller(tags=[xyz]) if exists
            if tags:
                if router.routes[route].tags is None: router.routes[route].tags = []
                router.routes[route].tags.extend(tags)
            routes.append(router.routes[route])

        return routes

    def include(self,
        module: Union[str, Callable],
        *,
        prefix: str = '',
        name: str = '',
        tags: Optional[List[str]] = None,
        options: Dict = {}
    ) -> List:
        """Alias to controller"""
        self.controller(module=module, prefix=prefix, name=name, tags=tags, options=options)

    def group(self, prefix: str = '', *,
        routes: Optional[List] = None,
        name: str = '',
        tags: Optional[List[str]] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
    ) -> Callable[[Decorator], Decorator]:
        """Route groups method and decorator"""

        # Convert auth helper param to middleware
        if middleware is None: middleware = []
        if auth: middleware.append(auth)

        # Convert scopes to Guard route middleware
        if scopes:
            middleware.append(Guard(scopes))

        # Get name
        if not name: name = prefix

        # Clean name
        if name:
            name = name.replace('/', '.')
            if name[-1] == '.': name = name[0:-1]  # Remove trailing .
            if name[0] == '.': name = name[1:]     # Remove beginning .

        def handle(routes):
            # Controllers return multiple routes as a List, so flatten everything into one List
            all_routes = []
            for route in routes:
                if type(route) == list:
                    all_routes.extend(route)
                else:
                    all_routes.append(route)

            # New routes with updated prefixes
            new_routes = []

            # Loop group routes and update prefixes
            for route in all_routes:

                # Remove old route from self.routes
                if route.name in self.routes:
                    self.routes.pop(route.name)

                # Strip global prefix if exists
                path = route.path
                if len(path) >= len(self.prefix) and path[0:len(self.prefix)] == self.prefix:
                    path = path[len(self.prefix):]

                rname = route.name
                if len(rname) >= len(self.name) and rname[0:len(self.name)] == self.name:
                    rname = rname[len(self.name) + 1:]

                full_path = prefix + path
                full_name = name + '.' + rname

                # Get route middleware based on parent child overrides
                (route_middleware, route.endpoint) = self._merge_route_middleware(middleware, route.middleware, route.endpoint)
                #dump(route, middleware, route.middleware)

                # Add new route with new group prefix and name
                # Because this is a polymorphic router (works for Web and API router)
                # The self.add methods will be different.  The actual route should mimic the self.add
                # parameters, so modify the route when calculated values and pass in as
                # self.add **kwargs
                if tags:
                    if route.tags is None: route.tags = []
                    route.tags.extend(tags)
                route.path = full_path
                route.name = full_name
                # Override group level autoprefix only if False to disable all, else use what the inside route used
                if autoprefix == False: route.autoprefix = autoprefix
                route.middleware = route_middleware
                del route.original_path
                del route.original_name

                # NO, this didn't handle polymorphic router.  self.add is different for each router.
                # new_route = self.add(
                #     path=full_path,
                #     endpoint=route.endpoint,
                #     methods=route.methods,
                #     name=full_name,
                #     autoprefix=autoprefix,
                #     #middleware=route.middleware or middleware  # Closest one to the route wins
                #     middleware=route_middleware
                # )
                new_route = self.add(**route)
                new_routes.append(new_route)

            # Return new routes for recursive nested groups
            return new_routes

        # Method access
        if routes: return handle(routes)

        # Decorator access
        def decorator(func: Decorator) -> Decorator:
            # Backup and clear existing routes
            all_routes = self._routes.clone()
            self._routes = Dict()

            # Get routes from the group method
            func()

            # Build routes list from group method
            routes = []
            for route in self._routes.values():
                routes.append(route)

            # Restore all routes besides the ones in the gruop method
            self._routes = all_routes.clone()

            # Add these routes to the proper group
            handle(routes)
            return func

        return decorator

    def _merge_route_middleware(self, parent_middleware: List, child_middleware: List, endpoint: Callable) -> Tuple[List, Callable]:
        """Merge parent route middleware into child, children parameters win in merge.

        If endpoing method is middleware, merge that also and return a new partial endpoint
        """

        # Merge helper
        def merge(parent_middleware, child_middleware):
            # Get middleware __init__ params
            inspection = inspect.signature(parent_middleware.__init__)
            params = [x for x in inspection.parameters]

            # Build params key value Dict
            parent_params = Dict()
            child_params = Dict()
            for param in params:
                # If param value is None do NOT add to kwargs or even the None will win in a merge
                if getattr(parent_middleware, param) is not None:
                    parent_params[param] = getattr(parent_middleware, param)
                if getattr(child_middleware, param) is not None:
                    child_params[param] = getattr(child_middleware, param)

            # Deep merge params
            kwargs = deep_merge(child_params, parent_params, merge_lists=True)

            # Instantiate new middleware with new params
            # We can't just set child_middleware new dict values becuase it has already fired up
            # its super.__init__, so changing values now does nothing.  Instead we replace it with an all
            # new instantiated Guard
            new_middleware = parent_middleware.__class__(**kwargs)
            return new_middleware

        # New merged middleware
        middlewares = []

        # Both parent and child have middleware. If both have the same middleware,
        # create a new middleware based on a merge of parameters where CHILD params WIN.
        if parent_middleware and child_middleware:
            for child_middleware in child_middleware:
                found = False
                for parent_middleware in parent_middleware:
                    if str(parent_middleware) == str(child_middleware):
                        found = True
                        break

                if not found:
                    # No matching parent middleware, use child
                    middlewares.append(child_middleware)
                else:
                    # Found matching parent and child middleware.  Create new middleware with merged parameters of the two
                    # Because of the break statement, the current parent_middleware variable is the match
                    middlewares.append(merge(parent_middleware, child_middleware))

        elif parent_middleware:
            # We have parent middleware, no children
            middlewares = copy(parent_middleware)

        elif child_middleware:
            # We have child middleware, no parent
            middlewares = copy(child_middleware)

        # Check each endpoints params and look at its default value
        # Compare that default value to our list of middleware
        # If they are the same, REMOVE the middleware from our list
        # or we will trigger the same middleware twice.  Once in a group/controller
        # and once on the actual function parameter (with FastAPI Depends) as well.
        final_middleware = []
        if middlewares:
            # Get endpoing method signature end parameters
            endpoint_signature = inspect.signature(endpoint)
            endpoint_params = endpoint_signature.parameters

            # Loop all merged parent/child middleware
            for middleware in middlewares:
                add = True

                # Loop each endpoint parameter
                for endpoint_param in endpoint_params.values():
                    # Endpoint middleware will be the default value of a parameter
                    endpoint_middleware = endpoint_param.default

                    # Compare the endpoint default value with the current middleware
                    # If they are the same, this endpint parameter is middleware and matches
                    # on of our previously defined merged middlewares.
                    # If we find a match do NOT add current middleware to final_middleware to eliminate
                    # duplicate middleware being run twice.
                    if str(middleware) == str(endpoint_middleware):
                        # Do NOT add this middleware to final_middleware
                        add = False

                        # Merge endpoint middleware to higher level matched middleware
                        # This will merge scopes and other parameters to provide scope hierarchy
                        new_endpoint_middleware = merge(middleware, endpoint_middleware)

                        # Create a partial endpoint method with our new middleware parameter injected
                        # Partials see https://levelup.gitconnected.com/changing-pythons-original-behavior-8a43b7d1c55d
                        endpoint = partial(endpoint, **{
                            endpoint_param.name: new_endpoint_middleware
                        })

                        # There will not be another endpoint param matching this one middleware, break
                        break

                if add:
                    # No endpoint middleware found, add current middleware
                    final_middleware.append(middleware)

        # Return new merged middleware and modified partial endpoint
        return (final_middleware, endpoint)

    def _merge_route_middlewareOLD(self, parent_middleware: List, child_middleware: List) -> List:
        """Merge parent route middleware into child, children parameters win in merge."""

        # Parent has no middleware to merge into child
        if not parent_middleware: return child_middleware

        # if not child_middleware:
        #     # Child has no middleware of its own, add parent level middleware to this route
        #     return copy(parent_middleware)

        # Merge helper
        def merge(parent_middleware, child_middleware):
            # Get middleware __init__ params
            inspection = inspect.signature(parent_middleware.__init__)
            params = [x for x in inspection.parameters]

            # Build params key value Dict
            parent_params = Dict()
            child_params = Dict()
            for param in params:
                # If param value is None do NOT add to kwargs or even the None will win in a merge
                if getattr(parent_middleware, param) is not None:
                    parent_params[param] = getattr(parent_middleware, param)
                if getattr(child_middleware, param) is not None:
                    child_params[param] = getattr(child_middleware, param)

            # Deep merge params
            kwargs = deep_merge(child_params, parent_params, merge_lists=True)

            # Instantiate new middleware with new params
            # We can't just set child_middleware new dict values becuase it has already fired up
            # its super.__init__, so changing values now does nothing.  Instead we replace it with an all
            # new instantiated Guard
            new_middleware = parent_middleware.__class__(**kwargs)
            return new_middleware


        # Both parent and child have middleware. If both have the same middleware,
        # create a new middleware based on a merge of parameters where CHILD params WIN.
        middlewares = []
        if child_middleware:
            for child_middleware in child_middleware:
                found = False
                for parent_middleware in parent_middleware:
                    if str(parent_middleware) == str(child_middleware):
                        found = True
                        break

                if not found:
                    # No matching controller middleware found
                    middlewares.append(child_middleware)
                else:
                    # Found matching controller and route middleware.  Create new middleware with merged parameters of the two
                    # Because of the break, the current parent_middleware variable is the match
                    middlewares.append(merge(parent_middleware, child_middleware))
        else:
            middlewares = copy(parent_middleware)

        return middlewares

    def _format_path_name(self, path: str, name: str, autoprefix: bool, methods: List):
        # Clean path
        if path and path[-1] == '/': path = path[0:-1]  # Remove trailing /
        if path and path[0] != '/': path = '/' + path   # Add beginning /

        # Get name
        no_name = False
        if not name:
            # We have no explicit name, derive name from path including params or its not unique
            no_name = True
            name = path.replace('{', '').replace('}', '')

        # Clean name
        name = name.replace('/', '.')
        if name and name[-1] == '.': name = name[0:-1]  # Remove trailing .
        if name and name[0] == '.': name = name[1:]     # Remove beginning .
        if not name: name = 'root'

        # Autoprefix path and name
        # Note that route "name" is for URL linking
        # This does NOT use the global URL prefix because that can change by the user running the package
        # Instead we use the actual package name for the name prefix
        full_path = path
        full_name = name
        if autoprefix:
            full_path = self.prefix + full_path
            full_name = self.name + '.' + full_name
        if not full_path: full_path = '/'

        # To prevent duplicates of same route for GET, POST, PUT, PATCH, DELETE
        # suffix the name with -POST, -PUT...  But do not suffix for GET
        # If this is a multi method function, only add POST-PUT combination of no name is forced
        methods = methods.copy()
        if 'GET' in methods: methods.remove('GET')
        if methods:
            if no_name or (no_name == False and len(methods) == 1):
                methods = sorted(methods)
                method_suffix = '-' + '-'.join(methods)
                if full_name[-len(method_suffix):] != method_suffix: full_name = full_name + method_suffix

        # Return newly formatted values
        return (path, full_path, name, full_name)


@uvicore.service()
class Routes(RoutesInterface):
    """Routes and Controller Class"""
    # middleware = None
    # auth = None
    # scopes = None

    @property
    def package(self) -> PackageInterface:
        return self._package

    def __init__(self, package: PackageInterface):
        self._package = package

    def _middleware(self):
        # Get class level middleware
        middlewares = []
        if self.auth: middlewares.append(self.auth)
        if self.middleware: middlewares.extend(self.middleware)
        if self.scopes: middlewares.append(Guard(self.scopes))

        return middlewares
        # if '__annotations__' in self.__class__.__dict__:
        #     for key, value in self.__class__.__annotations__.items():
        #         middlewares.append(getattr(self, key))
        # return middlewares





import uvicore
import json as JSON
from uvicore.console import command, argument, option
from uvicore import app, log
from uvicore.support.dumper import dd, dump


@command()
@option('--raw', is_flag=True, help='Show output without prettyprinter')
def list(raw: bool = False):
    """List all deep merged configs from all packages"""
    if not raw:
        log.header("Final Merged Configs")
        log.line()
        dump(uvicore.config)
    else:
        print(uvicore.config)


@command()
@argument('key', default='')
@option('--raw', is_flag=True, help='Show output without prettyprinter')
def get(key: str = None, raw: bool = False):
    """Get a config value by key"""
    if not raw:
        dump(uvicore.config.dotget(key))
    else:
        print(uvicore.config.dotget(key))



from abc import ABC, abstractmethod
from typing import Any, Dict, List

from uvicore.contracts import Application, Package, Dispatcher


class Provider(ABC):

    @property
    @abstractmethod
    def app(self) -> Application:
        """Uvicore application instance"""
        pass

    @property
    @abstractmethod
    def events(self) -> Dispatcher:
        """Event instance"""
        pass

    @property
    @abstractmethod
    def package(self) -> Package:
        """The current package class.  Not available in boot()"""
        pass

    @property
    @abstractmethod
    def app_config(self) -> Dict: pass

    @property
    @abstractmethod
    def package_config(self) -> Dict: pass

    @property
    def name(self) -> Dict: pass

    @abstractmethod
    def register(self) -> None: pass

    @abstractmethod
    def boot(self) -> None: pass

    @abstractmethod
    def bind(self,
        name: str,
        object: Any,
        *,
        factory: Any = None,
        kwargs: Dict = None,
        singleton: bool = False,
        aliases: List = []
    ) -> None:
        """Bind a concrete class to an IoC name"""
        pass

    @abstractmethod
    def configs(self, modules: List) -> None:
        """Register packages configs"""
        pass




    # @abstractmethod
    # def views(self, package: Package, paths: List) -> None:
    #     """Register packages view paths"""
    #     pass

    # @abstractmethod
    # def assets(self, package: Package, paths: List) -> None:
    #     """Register packages asset paths"""
    #     pass

    # @abstractmethod
    # def template(self, package: Package, options: Dict) -> None:
    #     """Register packages view templating options"""
    #     pass

    # @abstractmethod
    # def web_routes(self, package: Package, routes_class: Any) -> None:
    #     """Register packages web routes"""
    #     pass

    # @abstractmethod
    # def api_routes(self, package: Package, routes_class: Any) -> None:
    #     """Register packages API routes"""
    #     pass

    # @abstractmethod
    # def models(self, models: List[str]) -> None:
    #     """Register packages database ORM models"""
    #     pass

    # @abstractmethod
    # def tables(self, tables: List[str]) -> None:
    #     """Register packages database tables"""
    #     pass

    # @abstractmethod
    # def seeders(self, seeders: List[str]) -> None:
    #     """Register packages database seeders"""
    #     pass

    # @abstractmethod
    # def commands(self, package: Package, options: Dict) -> None:
    #     """Register packages CLI commands"""
    #     pass



import os
import sys
import uvicore
from uvicore.typing import Any, List, NamedTuple, Tuple, Dict, OrderedDict, Union
from uvicore.package import Package
from uvicore.contracts import Application as ApplicationInterface
from uvicore.contracts import Config as ConfigInterface
from uvicore.contracts import Package as PackageInterface
from uvicore.contracts import Server as ServerInterface
from uvicore.contracts import Template as TemplateInterface
from uvicore.support.collection import dotget
from uvicore.support.dumper import dd, dump
from uvicore.support.hash import md5
from uvicore.support.module import load, location
from uvicore.console import command_is

#import uvicore.foundation.events.app
from uvicore.foundation.events import app as events

try:
    from fastapi import FastAPI
    from starlette.applications import Starlette
except ImportError:  # pragma: nocover
    FastAPI = None  # type: ignore
    Starlette = None  # type: ignore


@uvicore.service('uvicore.foundation.application.Application',
    aliases=['Application', 'application', 'App', 'app'],
    singleton=True,
)
class Application(ApplicationInterface):
    """Application private class.

    Do not import from this location.
    Use the uvicore.app singleton global instead."""

    # Instance Variables
    @property
    def version(self) -> str:
        return self._version

    @property
    def debug(self) -> bool:
        return self._debug

    @property
    def perfs(self) -> List:
        return self._perfs

    @property
    def http(self) -> Union[Starlette, FastAPI]:
        return self._http

    # No, don't want duplicate entrypoints
    # @property
    # def config(self) -> ConfigInterface:
    #     return self._config

    @property
    def providers(self) -> OrderedDict[str, Dict]:
        return self._providers

    @property
    def registered(self) -> bool:
        return self._registered

    @property
    def booted(self) -> bool:
        return self._booted

    @property
    def is_console(self) -> bool:
        return self._is_console

    @property
    def is_http(self) -> bool:
        return self._is_http

    @property
    def packages(self) -> OrderedDict[str, PackageInterface]:
        return self._packages

    @property
    def running_config(self) -> OrderedDict:
        return self._running_config

    @property
    def path(self) -> str:
        return self._path

    @property
    def name(self) -> str:
        return self._name

    @property
    def main(self) -> Dict:
        return self._main


    def __init__(self):
        # Instance variables
        self._version = uvicore.__version__
        self._debug = False
        self._perfs = []
        self._http = None
        #self._config = None  # None until config provider registered
        self._providers = OrderedDict()
        self._registered = False
        self._booted = False
        self._is_console = False
        self._is_http = False
        self._packages = OrderedDict()
        self._running_config = OrderedDict()
        self._path = None
        self._name = None
        self._main = None


    def add_running_config(self, key: str, value):
        if type(value) == list:
            if not self._running_config.dotget(key): self._running_config.dotset(key, [])
            self._running_config.dotget(key).extend(value)
        else:
            self._running_config.dotset(key, value)


    def bootstrap(self, app_config: Dict, path: str, is_console: bool) -> None:
        """Bootstrap the uvicore application"""

        # Silently do not bootstrap multiple times
        if self.booted: return

        # App name and path
        self._path = path
        self._name = app_config.name
        self._main = app_config.main

        # Add to running config
        self.add_running_config('name', self.name)
        self.add_running_config('main', self.main)
        self.add_running_config('path', self.path)
        self.add_running_config('version', '0.0.0') # Fill in later from main running package config
        self.add_running_config('uvicore_version', self.version)

        # Detect if running in console (to register commands)
        # Ensure console is False even when running ./uvicore http serve
        self._is_console = is_console
        if command_is('http serve'): self._is_console = False
        self._is_http = not self.is_console

        # Detect debug flag from main app config
        self._debug = app_config.debug
        self.add_running_config('debug', self.debug)

        # Build recursive providers graph
        self._build_provider_graph(app_config)
        #self.add_running_config('providers', self.providers)

        # Failsafe if no http package, force console
        # This solves a ./uvicore http serve error if you don't have the http package
        #if 'uvicore.web' not in self.providers or 'uvicore.api' not in self.providers:
        if 'uvicore.http' not in self.providers:
            self._is_console = True
            self._is_http = False

        # Register all providers by calling each packages register() method
        # This is what builds self._packages with an actual Package Class objects
        # Uses self.providers and converts those into full Package objects
        self._register_providers(app_config)

        # Boot all providers by calling each packages boot() method
        self._boot_providers(app_config)

        # Add each final package to running_config, but only the basic details
        # self.add_running_config('packages', self.packages)
        for package in self.packages.values():
            if package.main:
                self.add_running_config('version', package.version)
            self.add_running_config('packages.[' + package.name + ']', {
                'name': package.name,
                'short_name': package.short_name,
                'vendor': package.vendor,
                'version': package.version,
                'main': package.main,
                'path': package.path,
                'registers': [k for (k,v) in package.registers.items() if v == True],
            })

        # Return application
        return self


    def package(self, package: str = None, *, main: bool = False, hint: str = None) -> PackageInterface:
        """Get package by name or by main running package"""
        if package:
            return self.packages.get(package)
            #return self.packages.dotget(package)
        elif main:
            return next(package for package in self.packages.values() if package.main == True)
        elif hint:
            return self.package('app1')
            #return self.packages.dotget(self.main)


    def perf(self, item) -> None:
        """Add entry to debug performance counter"""
        if self.debug:
            self.perfs.append(item)
            print(item)


    def _build_provider_graph(self, app_config: Dict) -> None:
        """Build recursive dependency graph of all packages"""

        def recurse(package: str, options: Dict):
            # Get the config/package.py config
            package_config = self._get_package_config(package, options)

            # Get the packages service provider dependencies
            services = package_config.get('dependencies') or {}

            # Recurse into each service dependency and find those dependencies from their config/package.py configs
            for service, details in services.items():
                recurse(service, details)

            # Add to providers OrderedDict. Notice this will OVERWRITE if new provider defined last.
            # This gives the perfect LAST provider WINS!  Also because this is an OrderedDict the
            # last provider will overrite the value but the defined ORDER will remain the
            # same.  Perfect!
            self._providers[package] = options

        # Graph starts with "main", this is the package acting as the main running app itself
        main = app_config.main
        if 'config' in main:
            # Allows main to specific a custom config location
            recurse(main.package, {
                'provider': main.provider,
                'config': main.config
            })
        else:
            recurse(main.package, {
                'provider': main.provider,
            })

        # Next we loop all app defined overrides.providers
        overrides = app_config.overrides.providers or {}
        for package, options in overrides.items():
            recurse(package, options)


    def _register_providers(self, app_config: Dict) -> None:
        """Register all providers by calling each ServiceProviders register() method"""

        for package_name, service in self.providers.items():
            # Example:
            # package_name = uvicore.configuration
            # service = {
            #   'provider': 'uvicore.configuration.services.Configuration',
            #   'config': 'uvicore.configuration.config.package.config',
            # }

            # Get the config/package.py config
            package_config = self._get_package_config(package_name, service)

            # Start a new package definition
            self._packages[package_name] = Package({
                'name': package_name,
                'short_name': package_name.split('.')[-1] if '.' in package_name else package_name,
                'vendor': package_name.split('.')[0],  # Works fine even if no .
                'version': package_config.version or '0.1.0',
                'main': True if package_name == self.main.package else False,
                'path': location(package_name),
            })

            # Instantiate the provider and call the register() method
            provider = load(service['provider']).object(
                app=self,
                name=package_name,
                package=None,  # Not available in register()
                app_config=app_config,
                package_config=package_config,
            )
            provider.register()

        # Complete registration
        self._registered = True
        #uvicore.events.dispatch('uvicore.foundation.events.app.Registered')
        #uvicore.events.dispatch(uvicore.foundation.events.app.Registered())
        #uvicore.events.dispatch('uvicore.foundation.events.app.Registered', {'test': 'test1'})
        events.Registered().dispatch()


    def _boot_providers(self, app_config: Dict) -> None:
        for package_name, service in self.providers.items():

            # Example:
            # package_name = uvicore.configuration
            # service = {'provider': 'uvicore.configuration.services.Configuration'}

            # Get the config/package.py config
            package_config = self._get_package_config(package_name, service)

            # Import the provider and call boot()
            provider = load(service['provider']).object(
                app=self,
                name=package_name,
                package=self.package(package_name),
                app_config=app_config,
                package_config=package_config,
            )
            provider.boot()

        # Complete booting
        self._booted = True
        #uvicore.events.dispatch('uvicore.foundation.events.app.Booted')
        #uvicore.events.dispatch(uvicore.foundation.events.app.Booted())
        #uvicore.events.dispatch('uvicore.foundation.events.app.Booted')
        events.Booted().dispatch()

    def _get_package_config(self, package: str, options: Dict) -> Dict:
        """Load packages configuration"""
        config_module = package + '.config.package.config'  # Default if not defined
        if 'config' in options: config_module = options['config']
        config = Dict()
        try:
            config = Dict(load(config_module).object)
        except:
            # Often we won't have any config for a package, if so return empty Dict
            pass
        return config

    # NO, don't want duplicates of everything everywhere, just import dumper
    # def dump(self, *args) -> None:
    #     dump(*args)

    # def dd(self, *args) -> None:
    #     dd(*args)


# IoC Class Instance
# **Not meant to be imported from here**.  Use the uvicore.app singleton global instead.
# Only here because uvicore bootstrap needs to import it without a service provider.
# By using the default bind and make feature of the IoC we can swap the implimentation
# at a high bootstrap level using our app configs 'bindings' dictionary.
# The only two classes that do this are Application and the event Dispatcher.
#Application: _Application = uvicore.ioc.make('Application', _Application, singleton=True, aliases=['App', 'app', 'application'])



from __future__ import annotations
import uvicore
from uvicore.typing import Any, Callable, List, Dict, Optional, Decorator
from uvicore.http.routing.router import Router
from uvicore.support.dumper import dump, dd
from uvicore.contracts import WebRoute as RouteInterface
from prettyprinter import pretty_call, register_pretty
from uvicore.http.routing.guard import Guard
from merge_args import _merge as merge_args


@uvicore.service()
class WebRouter(Router['WebRoute']):

    def get(self,
        # Common to both ApiRouter and WebRouter
        path: str,
        endpoint: Optional[Callable] = None,
        *,
        name: Optional[str] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
        inherits: Optional[Callable] = None,
    ):
        """Add a new HTTP GET route to the WebRouter

        :param path: URL path, beginning with /
        :param endpoint: None if decorator, else method to call when route is fired.
        :param name: Name of this route to consistently access with url('name') helper.
        :param autoprefix: Disable the name autoprefixer (autoprefix adds the appname. prefix).  Useful when overriding another packages route.
        :param middleware: List of route level middleware.
        :param auth: Shorthand for the middleware=[Guard(['scope'])].  Usage auth=Guard(['scope']).
        :param scopes: Shorthand for middleware=[Guard(['scope'])].  Usage scopes=['scope1', 'scope1'].
        """
        # Build parameters
        methods = ['GET']
        #params = {key:value for key, value in locals().items() if key != 'self'}
        params = locals()
        params.pop('self')

        # Pass to generic add method
        return self.add(**params)

    def post(self,
        # Common to both ApiRouter and WebRouter
        path: str,
        endpoint: Optional[Callable] = None,
        *,
        name: Optional[str] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
        inherits: Optional[Callable] = None,
    ):
        """Add a new HTTP POST route to the WebRouter

        :param path: URL path, beginning with /
        :param endpoint: None if decorator, else method to call when route is fired.
        :param name: Name of this route to consistently access with url('name') helper.
        :param autoprefix: Disable the name autoprefixer (autoprefix adds the appname. prefix).  Useful when overriding another packages route.
        :param middleware: List of route level middleware.
        :param auth: Shorthand for the middleware=[Guard(['scope'])].  Usage auth=Guard(['scope']).
        :param scopes: Shorthand for middleware=[Guard(['scope'])].  Usage scopes=['scope1', 'scope1'].
        """
        # Build parameters
        methods = ['POST']
        params = locals()
        params.pop('self')

        # Pass to generic add method
        return self.add(**params)

    def put(self,
        # Common to both ApiRouter and WebRouter
        path: str,
        endpoint: Optional[Callable] = None,
        *,
        name: Optional[str] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
        inherits: Optional[Callable] = None,
    ):
        """Add a new HTTP PUT route to the WebRouter

        :param path: URL path, beginning with /
        :param endpoint: None if decorator, else method to call when route is fired.
        :param name: Name of this route to consistently access with url('name') helper.
        :param autoprefix: Disable the name autoprefixer (autoprefix adds the appname. prefix).  Useful when overriding another packages route.
        :param middleware: List of route level middleware.
        :param auth: Shorthand for the middleware=[Guard(['scope'])].  Usage auth=Guard(['scope']).
        :param scopes: Shorthand for middleware=[Guard(['scope'])].  Usage scopes=['scope1', 'scope1'].
        """
        # Build parameters
        methods = ['PUT']
        params = locals()
        params.pop('self')

        # Pass to generic add method
        return self.add(**params)

    def patch(self,
        # Common to both ApiRouter and WebRouter
        path: str,
        endpoint: Optional[Callable] = None,
        *,
        name: Optional[str] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
        inherits: Optional[Callable] = None,
    ):
        """Add a new HTTP PATCH route to the WebRouter

        :param path: URL path, beginning with /
        :param endpoint: None if decorator, else method to call when route is fired.
        :param name: Name of this route to consistently access with url('name') helper.
        :param autoprefix: Disable the name autoprefixer (autoprefix adds the appname. prefix).  Useful when overriding another packages route.
        :param middleware: List of route level middleware.
        :param auth: Shorthand for the middleware=[Guard(['scope'])].  Usage auth=Guard(['scope']).
        :param scopes: Shorthand for middleware=[Guard(['scope'])].  Usage scopes=['scope1', 'scope1'].
        """
        # Build parameters
        methods = ['PATCH']
        params = locals()
        params.pop('self')

        # Pass to generic add method
        return self.add(**params)

    def delete(self,
        # Common to both ApiRouter and WebRouter
        path: str,
        endpoint: Optional[Callable] = None,
        *,
        name: Optional[str] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
        inherits: Optional[Callable] = None,
    ):
        """Add a new HTTP DELETE route to the WebRouter

        :param path: URL path, beginning with /
        :param endpoint: None if decorator, else method to call when route is fired.
        :param name: Name of this route to consistently access with url('name') helper.
        :param autoprefix: Disable the name autoprefixer (autoprefix adds the appname. prefix).  Useful when overriding another packages route.
        :param middleware: List of route level middleware.
        :param auth: Shorthand for the middleware=[Guard(['scope'])].  Usage auth=Guard(['scope']).
        :param scopes: Shorthand for middleware=[Guard(['scope'])].  Usage scopes=['scope1', 'scope1'].
        """
        # Build parameters
        methods = ['DELETE']
        params = locals()
        params.pop('self')

        # Pass to generic add method
        return self.add(**params)

    def add(self,
        # Common to both ApiRouter and WebRouter
        path: str,
        endpoint: Optional[Callable] = None,
        methods: List[str] = ['GET'],
        *,
        name: Optional[str] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
        inherits: Optional[Callable] = None,
    ) -> Callable[[Decorator], Decorator]:
        """Add a new HTTP method route to the WebRouter

        :param path: URL path, beginning with /
        :param endpoint: None if decorator, else method to call when route is fired.
        :param name: Name of this route to consistently access with url('name') helper.
        :param autoprefix: Disable the name autoprefixer (autoprefix adds the appname. prefix).  Useful when overriding another packages route.
        :param middleware: List of route level middleware.
        :param auth: Shorthand for the middleware=[Guard(['scope'])].  Usage auth=Guard(['scope']).
        :param scopes: Shorthand for middleware=[Guard(['scope'])].  Usage scopes=['scope1', 'scope1'].
        """
        # Convert auth and scope helper param to middleware
        if middleware is None: middleware = []
        if auth: middleware.append(auth)
        if scopes: middleware.append(Guard(scopes))

        # Clean path and name
        (path, full_path, name, full_name) = self._format_path_name(path, name, autoprefix, methods)

        def handle(endpoint):
            # Merge function arguments if inheriting
            if inherits: endpoint = merge_args(inherits, endpoint)

            # Create route SuperDict
            route = WebRoute({
                # Common to both ApiRouter and WebRouter
                'path': full_path,
                'name': full_name,
                'methods': methods,
                'endpoint': endpoint,
                'middleware': middleware,
                'autoprefix': autoprefix,
                'original_path': path,
                'original_name': name,
            })
            # Full name already contains -POST, -PUT...for uniqueness across methods
            self.routes[full_name] = route
            return route

        # Method access
        if endpoint: return handle(endpoint)

        # Decorator access
        def decorator(func: Decorator) -> Decorator:
            handle(func)
            return func
        return decorator


@uvicore.service()
class WebRoute(RouteInterface):
    """Route superdict definition"""
    pass


@register_pretty(WebRoute)
def pretty_entity(value, ctx):
    """Custom pretty printer for my SuperDict"""
    # This printer removes the class name uvicore.types.Dict and makes it print
    # with a regular {.  This really cleans up the output!

    # SuperDict are printed as Dict, but this Package SuperDict should
    # be printed more like a class with key=value notation, so use **values
    return pretty_call(ctx, 'WebRoute', **value)
















# # Example of using FastAPI router even for Web routes
# @uvicore.service()
# class WebRouterX(ApiRouter):

#     def get(
#         self,
#         path: str,
#         *,
#         #response_model: Optional[Type[Any]] = None,
#         status_code: int = 200,
#         #tags: Optional[List[str]] = None,
#         dependencies: Optional[Sequence[params.Depends]] = None,
#         #summary: Optional[str] = None,
#         #description: Optional[str] = None,
#         #response_description: str = "Successful Response",
#         #responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
#         #deprecated: Optional[bool] = None,
#         #operation_id: Optional[str] = None,
#         #response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
#         #response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
#         #response_model_by_alias: bool = True,
#         #response_model_exclude_unset: bool = False,
#         #response_model_exclude_defaults: bool = False,
#         #response_model_exclude_none: bool = False,
#         #include_in_schema: bool = False,  # Uvicore change
#         #response_class: Type[Response] = Default(JSONResponse),
#         name: Optional[str] = None,
#         #callbacks: Optional[List[BaseRoute]] = None,
#     ) -> Callable[[DecoratedCallable], DecoratedCallable]:
#         return super().get(
#             path=path,
#             #response_model=response_model,
#             status_code=status_code,
#             #tags=tags,
#             dependencies=dependencies,
#             #summary=summary,
#             #description=description,
#             #response_description=response_description,
#             #responses=responses,
#             #deprecated=deprecated,
#             #operation_id=operation_id,
#             #response_model_include=response_model_include,
#             #response_model_exclude=response_model_exclude,
#             #response_model_by_alias=response_model_by_alias,
#             #response_model_exclude_unset=response_model_exclude_unset,
#             #response_model_exclude_defaults=response_model_exclude_defaults,
#             #response_model_exclude_none=response_model_exclude_none,
#             include_in_schema=False,  # Don't show in openapi docs
#             response_class=response.HTML,
#             name=name,
#             #callbacks=callbacks,
#         )



# # Example of just using the entier Starlette Router without abstraction
# @uvicore.service()
# class WebRouterXX(_StarletteRouter):

#     # Add a GET route decorator to starlette
#     def get(self,
#         path: str,
#         name: str = None,
#     ) -> Callable:
#         return self.route(
#             path=path,
#             name=name,
#             methods=['GET'],
#             include_in_schema=False,
#         )

#     # Add a POST route decorator to starlette
#     def post(self,
#         path: str,
#         name: str = None,
#     ) -> Callable:
#         return self.route(
#             path=path,
#             name=name,
#             methods=['POST'],
#             include_in_schema=False,
#         )

#     # Add a PUT route decorator to starlette
#     def put(self,
#         path: str,
#         name: str = None,
#     ) -> Callable:
#         return self.route(
#             path=path,
#             name=name,
#             methods=['PUT'],
#             include_in_schema=False,
#         )

#     # Add a PATCH route decorator to starlette
#     def patch(self,
#         path: str,
#         name: str = None,
#     ) -> Callable:
#         return self.route(
#             path=path,
#             name=name,
#             methods=['PATCH'],
#             include_in_schema=False,
#         )

#     # Add a DELETE route decorator to starlette
#     def delete(self,
#         path: str,
#         name: str = None,
#     ) -> Callable:
#         return self.route(
#             path=path,
#             name=name,
#             methods=['DELETE'],
#             include_in_schema=False,
#         )

#     # Add a OPTIONS route decorator to starlette
#     def options(self,
#         path: str,
#         name: str = None,
#     ) -> Callable:
#         return self.route(
#             path=path,
#             name=name,
#             methods=['OPTIONS'],
#             include_in_schema=False,
#         )

#     # Add a HEAD route decorator to starlette
#     def head(self,
#         path: str,
#         name: str = None,
#     ) -> Callable:
#         return self.route(
#             path=path,
#             name=name,
#             methods=['HEAD'],
#             include_in_schema=False,
#         )

#     # Add a TRACE route decorator to starlette
#     def trace(self,
#         path: str,
#         name: str = None,
#     ) -> Callable:
#         return self.route(
#             path=path,
#             name=name,
#             methods=['TRACE'],
#             include_in_schema=False,
#         )

#     def include_router(self, router: "WebRouter") -> None:
#         # Starlette does not have a _router.include_router  like FastAPI
#         # This manually adds each route in the router to starlette router
#         # Used in the Class Based Controllers _init_cbv code
#         for route in router.routes:
#             self.routes.append(route)




# # Example of making my own router abstraction
# class WebRouterXXX(RouterInterface):

#     @property
#     def router(self) -> _StarletteRouter:
#         return self._router

#     @property
#     def routes(self) -> List[BaseRoute]:
#         return self._router.routes

#     @property
#     def on_startup(self) -> None:
#         return self._router.on_startup

#     @property
#     def on_shutdown(self) -> None:
#         return self._router.on_shutdown

#     def __init__(self):
#         # Fireup Starlette Router
#         self._router = _StarletteRouter()

#     def get(self,
#         path: str,
#         name: str = None,
#     ) -> Callable:
#         return self._router.route(
#             path=path,
#             name=name,
#             methods=['GET']
#         )

#     def post(self,
#         path: str,
#         name: str = None,
#     ) -> Callable:
#         return self._router.route(
#             path=path,
#             name=name,
#             methods=['POST']
#         )

#     # Not needed - was used for Class Based Controllers
#     # def include_router(self, router: "WebRouter") -> None:
#     #     # Starlette does not have a _router.include_router  like FastAPI
#     #     # This manually adds each route in the router to starlette router
#     #     for route in router.routes:
#     #         self._router.routes.append(route)


# # IoC Class Instance
# #WebRouter: RouterInterface = uvicore.ioc.make('WebRouter', _WebRouter)

# # Public API for import * and doc gens
# #__all__ = ['WebRouter', '_WebRouter']



from __future__ import annotations
from typing import Optional, List
from uvicore.contracts import Model as ModelInterface


# Model interfaces, though redundant, are used for proper type hinting code intellisense
class Group(ModelInterface):
    id: Optional[int]
    name: str


# Import relations after model to avoid circular dependencies



from __future__ import annotations

import operator as operators
import os

from collections import OrderedDict as ODict
from copy import deepcopy
from typing import Any, Dict, Generic, List, OrderedDict, Tuple, TypeVar, Union, Callable
from uvicore.support.hash import sha1

import sqlalchemy as sa
from sqlalchemy.sql import quoted_name
from sqlalchemy.sql.expression import BinaryExpression

import uvicore
from uvicore.contracts import OrmQueryBuilder as BuilderInterface
from uvicore.database.builder import QueryBuilder, Join, Query
from uvicore.orm.fields import (BelongsTo, BelongsToMany, Field, HasMany,
                                HasOne, MorphMany, MorphOne, MorphToMany)
from uvicore.orm.fields import Relation
from uvicore.support.collection import getvalue
from uvicore.support.dumper import dd, dump

B = TypeVar("B")  # Builder Type (DbQueryBuilder or OrmQueryBuilder)
E = TypeVar("E")  # Entity Model


@uvicore.service()
class OrmQueryBuilder(Generic[B, E], QueryBuilder[B, E], BuilderInterface[B, E]):
    """ORM Query Builder"""

    def __init__(self, entity: E):
        self.entity = entity
        super().__init__()

        # Not all models require tables (databaseless models)
        if entity.table is not None:
            self.query.table = entity.table
        else:
            self.query.table = entity.tablename

    @property
    def log(self):
        return uvicore.log.name('uvicore.orm')

    def _where_dict(self, parent_method: Callable, column: str, operator: str = None, value: Any = None):
        # Not sure I want this code if I can't do where AND and where OR etc...
        # Maybe revisit later.
        found_dict_where = False
        if type(column) == str:
            # Swap operator and value
            if not value: value = operator; operator = '='
            if '.' in column:
                parts = column.split(".")
                if len(parts) == 2:
                    fieldname = parts[0]
                    fieldvalue = parts[1]
                    field = self.entity.modelfields.get(fieldname)  # Don't use modelfield() as it throws exception
                    relation = field.relation.fill(field)
                    if fieldvalue not in relation.entity.modelfields.keys():
                        dict_key = getvalue(relation, 'dict_key')
                        dict_value = getvalue(relation, 'dict_value')
                        if dict_key and type(dict_value) == str:
                            found_dict_where = True

        if found_dict_where:
            # Dict where found, convert to proper where based on dict_key
            return parent_method([
                (fieldname + '.' + dict_key, '=', fieldvalue),
                (fieldname + '.' + dict_value, operator, value)
            ])
            pass
        else:
            # Regular where, pass to parent (builder.py) where
            return parent_method(column, operator, value)

    def where(self, column: Union[str, BinaryExpression, List[Union[Tuple, BinaryExpression]]], operator: str = None, value: Any = None) -> B[B, E]:
        # Custom where just for OrmQueryBuilder only to check for dict_key and dict_value type wheres
        # This is very limited and does not work with ORs or multiple ANDs.  Why not multiple ANDS?  Try querying posts join attributes and
        # where key=x and value=y and key=a and value=b and see what happens (you get nothing).  This is the nature of polymorphic one-to-many
        # So you can where on one attribute with one or more (in) values, but you cannot where to TWO attributes with AND
        # I also cannot do complex ANDs with ORs mixed around yet.
        # Like this is not yet possible in the ORM.  It can only do all ANDs then a final AND (asdf or asdf or asdf)
        # WHERE
        # (attributes.key = 'post1-test1' and attributes.value = 'value for post1-test1')
        # or
        # (attributes.key = 'post2-test1' and attributes.value = 'value for post2-test1')
        return self._where_dict(super().where, column, operator, value)

    def include(self, *args) -> B[B, E]:
        """Include child relation models"""
        # import inspect
        # x = inspect.currentframe()
        # y = inspect.getouterframes(x, 1)
        # context = y[1].code_context
        # #dump(context)
        #args = ['creator.contact']

        # Convert List to Args
        if len(args) == 1 and type(args[0]) == list: args = args[0]

        # Loop each arg and add to query.includes
        for include in args:
            self.query.includes.append(include)
        return self

    def filter(self, column: Union[str, BinaryExpression, List[Union[Tuple, BinaryExpression]]], operator: str = None, value: Any = None) -> B[B, E]:
        """Filter child relationship by this AND clause"""
        # Filters are for Many relations only
        if type(column) == str or type(column) == sa.Column:
            # A single filter as a string or actual SQLAlchemy Column
            # String default =
            #   .filter('column', 'value')
            # String explicit operator:
            #   .filter('column, '=', 'value')
            # SA Column default =
            #   .filter(table.column, 'value')
            # SA Column explicit operator:
            #   .filter(table.column, '=', 'value')

            # Swap operator and value
            if not value: value = operator; operator = '='
            self.query.filters.append((column, operator.lower(), value))
        elif type(column) == list:
            # Multiple filters in one as a List[Tuple] or List[BinaryExpression]
            for filter in column:
                # Recursivelly add Tuple filters
                if type(filter) == tuple:
                    # String
                    #   .filter([('column', 'value'), ('column', '=', 'value')])
                    # SA Column
                    #   .filter([(table.column, 'value'), (table.column, '=', 'value)])
                    if len(filter) == 2:
                        self.filter(filter[0], '=', filter[1])
                    else:
                        self.filter(filter[0], filter[1], filter[2])
                else:
                    # SQLAlchemy Binary Expression
                    # .filter([table.column == 'value', table.column >= 'value'])
                    self.filter(filter)
        else:
            # Direct SQLAlchemy expression
            # .filter(table.column == 'value' and table.column >= 'value')
            self.query.filters.append(column)
        return self

    def or_filter(self, filters: List[Union[Tuple, BinaryExpression]]) -> B[B, E]:
        """Filter child relationship by this OR clause"""
        # Or filter must be a list of tuple or BinaryExpression as it requires at least 2 statements
        # .or_filter([('column', 'value'), ('column', '=', 'value')])
        # .or_filter([table.column == value, table.column == value])
        or_filters: List[Tuple] = []
        for filter in filters:
            if type(filter) == tuple:
                if len(filter) == 2:
                    or_filters.append((filter[0], '=', filter[1]))
                else:
                    or_filters.append((filter[0], filter[1].lower(), filter[2]))
            else:
                # SQLAlchemy Binary Expression
                or_filters.append(filter)
        self.query.or_filters.extend(or_filters)
        return self

    def sort(self, column: Union[str, List[str], List[Tuple], Any], order: str = 'ASC') -> B[B, E]:
        """Sort Many relations only"""
        # This will not work as binary expression, becuase relation name is often
        # different than colums tablename
        if type(column) == str or type(column) == sa.Column:
            # A single sort as a string
            self.query.sort.append((column, order.upper()))
        else:
            # Multiple sort in one as List or List[Tuple]
            for sort in column:
                if type(sort) == tuple:
                    if len(sort) == 2:
                        self.sort(sort[0], sort[1])
                    else:
                        self.sort(sort[0], 'ASC')
                else:
                    self.sort(sort, 'ASC')
        return self

    def key_by(self, field: str) -> B[B, E]:
        """Key results as a Dictionary by this column"""
        self.query.keyed_by = field
        return self

    def show_writeonly(self, fields: List = None):
        if fields is None:
            self.query.show_writeonly = True
        else:
            self.query.show_writeonly = fields
        return self

    def sql(self, method: str = 'select', queries: List = None) -> str:
        """Get all SQL queries involved in this ORM statement"""
        if queries is None: queries = self._build_orm_queries(method)
        sqls = {}
        for query in queries:
            #sqls += '-- ' + query.get('name').upper() + ':' + os.linesep + query.get('sql') + os.linesep + os.linesep
            sqls[query.get('name')] = query.get('sql').replace('\n', '')
        return sqls

    def queries(self, method: str = 'select') -> List:
        """Get all queries involved in this ORM statement"""
        return self._build_orm_queries('select')

    async def find(self, pk_value: Union[int, str] = None, **kwargs) -> Union[E, None]:
        """Execute query by primary key or custom column and return first row found"""
        if pk_value:
            column = self._pk()
            value = pk_value
        elif kwargs:
            column = [x for x in kwargs.keys()][0]
            value = [x for x in kwargs.values()][0]

        # Ensure some incompatible builder methods are reset
        self.query.wheres = []
        self.query.or_wheres = []
        self.query.order_by = []
        self.query.keyed_by = None

        # Add in where on PK
        self.where(column, value)

        # Get List of Entities based on query results
        entities = await self.get()

        # Return one record or None
        if entities: return entities[0]
        return None

    async def get(self) -> Union[List[E], Dict[str, E]]:
        """Execute a select query and return all rows found"""

        # Get this models connection configuration
        #connection = uvicore.db.connection(self._connection())
        #backend = connection.backend
        #if backend != 'sqlalchemy':
            # Using a custom, non sqlalchemy backend.  Could be an API, CSV, JSON...
            # Load up the custom backend and fire off the get() executor
            #dump('not sqlalchemy')

        # Build SQLAlchemy select queries
        queries = self._build_orm_queries('select')
        #dump(queries)

        self.log.nl().header('Queries')
        self.log.dump(queries)

        self.log.header('Raw SQL Queries')
        self.log.info(self.sql('select', queries))

        # Get hook?  Experimental
        if hasattr(self.entity, 'get'):
            return await self.entity.get(queries)

        # Detect caching
        cache = self.query.cache
        if cache:
            prefix = 'uvicore.orm/'
            if cache.get('key') is None:
                # No cache name specified, automatically build unique based on queries
                query_hash = ''
                for query in queries:
                    if query['name'] == 'main':
                        query_hash = query.get('query').hash(
                            hash_type='sha1',
                            package='uvicore.orm',
                            entity=self.entity,
                            connection=self._connection()
                        )
                        break
                cache['key'] = prefix + query_hash
                #dump(query_hash)
            else:
                cache['key'] = prefix + cache.get('key')

        if cache and await uvicore.cache.has(cache.get('key')):
            # Cache found, use cached results
            #dump('ORM FROM CACHE')
            entities = await uvicore.cache.get(cache.get('key'))
        else:
            # Execute each query
            results = None
            main_query = None
            has_many = {}
            for query in queries:
                if query.get('name') == 'main':
                    main_query = query.get('query')
                    results = await self.entity.fetchall(query.get('saquery'))
                else:
                    has_many[query.get('name')] = await self.entity.fetchall(query.get('saquery'))

            # Convert results to List of entities
            entities = self._build_orm_results(main_query, results, has_many)

            # Add to cache if desired
            if cache: await uvicore.cache.put(cache.get('key'), entities, seconds=cache.get('seconds'))

        # Return List of Entities
        return entities

    async def delete(self) -> None:
        """Execute delete query"""

        # Build SQLAlchemy delete query
        query, saquery = self._build_query('delete', self.query.copy())

        # Execute query
        await self.entity.execute(saquery)

    async def update(self, **kwargs) -> None:
        """Execute update query"""

        # Build SQLAlchemy delete query
        query, saquery = self._build_query('update', self.query.copy())

        # Add in values
        saquery = saquery.values(**kwargs)

        # Execute query
        await self.entity.execute(saquery)

    def _build_orm_queries(self, method: str) -> List:
        # Different than the single _build_query in the DB Builder
        # This one is for ORM only and build multiple DB queries from one ORM query.
        queries = []

        # First query
        query = self.query.copy()

        # Build relation (join) queries
        self._build_orm_relations(query)

        # Add all columns from main model
        query.selects = self.entity.selectable_columns(show_writeonly=self.query.show_writeonly)

        # Add all selects where any nested relation is NOT a *Many
        relation: Relation
        for relation in query.relations.values():
            if not relation.contains_many(query.relations):
                # Don't use the relation.entity table to get columns, use the join aliased table
                table = self._get_join_table(query, alias=relation.name)
                columns = relation.entity.selectable_columns(table, show_writeonly=self.query.show_writeonly)
                for column in columns:
                    query.selects.append(column.label(quoted_name(relation.name + '__' + column.name, True)))

        # Build first query
        saquery = None
        if query.table is not None:
            query, saquery = self._build_query(method, query)
        queries.append({
            'name': 'main',
            'query': query,
            'saquery': saquery,
            'sql': str(saquery) if saquery is not None else '',
        })


        # So we have our first query perfect
        # Now we need to build a second or more queries for all *Many relations
        relation: Relation
        for relation in query.relations.values():
            # Only handle *Many relations
            if not relation.is_many(): continue

            # Relation __ name converted to dot name
            rel_dot = relation.name.replace('__', '.')

            # New secondary relation query
            query2 = self.query.copy()

            # Build ORM Relations but force HasMany joins to INNER JOIN
            self._build_orm_relations(query2)

            # Only if Many-To-Many add in the main tables pivot ID
            if type(relation) == BelongsToMany or type(relation) == MorphToMany:
                # Bug found while grabbing the table by name.  What if we join the same table twice (though different tables of course)
                # There will be 2 joins with the same table name and _get_join_table will grab the first one in error.
                # Instead we need to use the table alias name + __pivot to grab the proper unique table
                #join_table = self._get_join_table(query2, relation.join_tablename)  # No, using table name will clash if joining multiples of the same table,
                join_table = self._get_join_table(query2, alias=relation.name + '__pivot')  # Relation.name is the alias name which should match the unique join table even if multiples
                query2.selects.append(
                    getattr(join_table.c, relation.left_key).label(
                        quoted_name(relation.name + '__' + relation.left_key, True)
                    )
                )

            # Set selects to only those in the related table
            table = self._get_join_table(query2, alias=relation.name)
            columns = relation.entity.selectable_columns(table, show_writeonly=self.query.show_writeonly)
            for column in columns:
                query2.selects.append(column.label(quoted_name(relation.name + '__' + column.name, True)))

            # Add in selects for any *One sub_relations
            for sub_relation in query2.relations.values():
                if relation.name + '__' not in sub_relation.name: continue
                if sub_relation.contains_many(query2.relations, skip=relation.name.split('__')): continue
                table = self._get_join_table(query2, alias=sub_relation.name)
                columns = sub_relation.entity.selectable_columns(table, show_writeonly=self.query.show_writeonly)
                for column in columns:
                    query2.selects.append(column.label(quoted_name(sub_relation.name + '__' + column.name, True)))

            # Remove any wheres for this relation.  Why?  Because the relation
            # where is applied to the main query not to relations.  We still
            # show all relations for any main query item
            # If you want to filter relations, use .filter() instead
            new_wheres = []
            for where in query2.wheres:
                # This also matches any sub-relations uder the where and removes those too
                # BROKEN if using SQLAlchemy binary expressions
                # .where(section.name, 'Production')
                if '.' in where[0]:
                    # Perhaps the ORM should not be "hybrid" and always use string dotnotation?
                    where_entity = '.'.join(where[0].split('.')[0:-1]).lower()
                    if rel_dot[0:len(where_entity)] == where_entity:
                        # Found where for this relation or a sub-relation
                        # Notice I continue, therefore I am removing this relations where
                        continue
                new_wheres.append(where)
            query2.wheres = new_wheres

            # Add .filter() as .where()
            query2.wheres.extend(query2.filters)

            # Add .or_filter() as .or_where()
            query2.or_wheres.extend(query2.or_filters)

            # Add where to show only joined record that were found
            # Cant use INNER JOIN instead because it would limit further sub many-to-many
            query2.wheres.append((relation.name + '.' + relation.entity.pk, '!=', None))

            # Swap .sort() that apply to this relation as an order_by
            # This WIPES out any .order_by as they do not apply to relation queries
            new_sorts = []
            for sort in query2.sort:
                (sort_column, sort_order) = sort
                if '.'.join(sort_column.split('.')[0:-1]).lower() == rel_dot.lower():
                    # Found sort for this relation
                    new_sorts.append(sort)
            query2.sort = new_sorts
            query2.order_by = query2.sort

            # Build secondary relation query
            query2, saquery2 = self._build_query(method, query2)
            queries.append({
                'name': relation.name,
                'query': query2,
                'saquery': saquery2,
                'sql': str(saquery2),
            })

        # Return all queries
        return queries

    def _build_orm_relations(self, query: Query) -> None:
        if not query.includes: return

        def extract(field, value):
            entity = None
            foreign = 'id'
            local = field.name + '_id'
            if type(value) == tuple:
                entity = value[0]
                if len(value) >= 2:
                    foreign = value[1]
                if len(value) == 3:
                    local = value[2]
            else:
                entity = value
            return entity, foreign, local

        # Loop each 'include' string and JOIN in proper child relation tables
        relations: OrderedDict[str, Relation] = ODict()
        for include in query.includes:

            # Includes are in dotnotation form (creator.info, posts.comments etc...)
            # Each part (.) is a table and nested child tables.  Split out each part.
            parts = [include]
            if '.' in include: parts = include.split('.')

            # Each separate include (including all parts in that include) walk down in entities
            # starting with the main model entity (self.entity).  So for each include, start at the main entity.
            entity = self.entity

            # Loop each "include part" as each are nested tables with joins
            parts_added = []
            for part in parts:

                # Get field for this "include part".  Remember entity here is not self.entity, but a walkdown
                # based on each part (separated by dotnotation)
                field: Field = entity.modelfields.get(part)  # Don't use modelfield() as it throws exception

                # Field not found, include string was a typo
                if not field: continue

                # Field has a column entry, which means its NOT a relation
                if field.column: continue

                # Field is not an actual relation
                if not field.relation: continue

                # Keep track of parts we have already completed
                parts_added.append(part)

                # Relation name is a __ join of all parts walked thus far
                relation_name = '__'.join(parts_added)

                # Get actual relation object
                relation = field.relation.fill(field)

                # Set a new name based on relation_name dot notation for nested relations
                relation.name = relation_name

                # Add relation to List only once
                if relation_name not in relations:
                    # We have to deepcopy a relationship becuase if we have owner and creator
                    # and both of those user models have a "Contact" model, that contact model is a
                    # single instance.  We want separate instances of each relationship
                    relations[relation_name] = deepcopy(relation)

                    # Alias the Joined Table (so we can join the same table multiple times if needed, like owner and creator)
                    # Alias is always the relation_name, we'll just make it doubly clear with its own variable
                    alias=relation_name

                    # Get the table we are joining on (the left table)
                    # If this is the first loop in a "include part", the left_table is the entity table (still walking down)
                    # After that its the previous aliased table
                    left_table = entity.table
                    if len(parts_added) > 1:
                        # Get the table from the join based on all but last part of relation_name
                        left_tablename = '__'.join(relation_name.split('__')[0:-1])
                        left_table = self._get_join_table(query, alias=left_tablename)

                    # Get the join (right) table
                    join_table = relation.entity.table
                    if join_table.name != alias:
                        join_table = sa.alias(relation.entity.table, name=alias)

                    # Many-To-Many
                    if type(relation) == BelongsToMany or type(relation) == MorphToMany:
                        # We have to join 2 tables in a many-to-many
                        # First the relation table (sometimes called intermediate or pivot table)
                        # Second the actual related table
                        # No need to alias *Many joins like we do *One joins because it will never be done twice

                        # Get our pivot (middle) join table
                        # Needs to be an auto incremented ALIAS in case we join the same table twice (user creator and owner...)
                        pivot_join_table = sa.alias(relation.join_table) # Notice no name=alias, just let it append _1, _2... for pivot

                        # Join the *Many Pivot Table
                        left = self._column(getattr(left_table.c, entity.pk))
                        right = self._column(getattr(pivot_join_table.c, relation.left_key))
                        join = Join(
                            table=pivot_join_table,
                            tablename=relation.join_tablename,
                            left=left,
                            right=right,
                            onclause=left.sacol == right.sacol,
                            alias=alias + '__pivot',
                            method='outerjoin'
                        )
                        query.joins.append(join)

                        # Now join the relation/intermediate/pivot table with the related table
                        left = self._column(getattr(pivot_join_table.c, relation.right_key))
                        right = self._column(getattr(join_table.c, relation.entity.pk))
                        join = Join(
                            #table=relation.entity.table,
                            table=join_table,
                            tablename=relation.entity.tablename,
                            left=left,
                            right=right,
                            onclause=left.sacol == right.sacol,
                            alias=alias,
                            method='outerjoin'
                        )
                        query.joins.append(join)

                    else:
                        # These One-To-One or One-To-Many on the BelongsTo side can be joined
                        # Multiple times.  For exampel post creator and post owner both join User table
                        # So they require aliases

                        # Debug dump which tables are which
                        #dump('relation_name: ' + relation_name + ' - main table: ' + left_table.name + ' - join_table: ' + join_table.name + ' - alias: ' + alias)

                        # Map relation.local_key and foreign_key Field names to column names
                        #local_key = relation.entity.mapper(relation.local_key).column()
                        #foreign_key = relation.entity.mapper(relation.foreign_key).column()
                        local_key = relation.local_key
                        foreign_key = relation.foreign_key

                        # Join condition columns
                        left = self._column(getattr(left_table.c, local_key))
                        right = self._column(getattr(join_table.c, foreign_key))

                        # Onclause, default for most relations
                        onclause = left.sacol == right.sacol

                        # Onclause override for Polymorphic relations
                        if type(relation) == MorphOne or type(relation) == MorphMany:
                            # In polymorphic, add the entity type to the onclause using and_
                            # Purpose is to get this JOIN ON AND CLAUSE
                            # ...JOIN attributes ON attributable_type = 'posts' AND id = attributable_id

                            # Get the ploy_type column (ex: attributable_type)
                            poly_type = self._column(getattr(join_table.c, relation.foreign_type))

                            # That 'posts' is the actual posts table.  But sometimes the table is an Alias
                            # like 'post' (singular).  So we need to use _get_tablename to get the REAL tablename even an Alias

                            # poly_type.sacol is 'attributable_type'
                            # self.get_tablename(left.table) is 'posts'
                            # left.sacol is 'id'
                            # right.sacol is 'attributable_id'
                            onclause = sa.and_(poly_type.sacol == self._get_tablename(left.table), left.sacol == right.sacol)

                        # Append new Join
                        join = Join(
                            table=join_table,
                            tablename=str(join_table.name),
                            left=left,
                            right=right,
                            onclause=onclause,
                            alias=alias,
                            method='outerjoin'
                        )
                        query.joins.append(join)


                    # All other types of relations
                    # elif type(relation) == HasMany:
                    #     # The One-To-Many on the HasMany side will never be joined
                    #     # twice and therefore do not need aliases
                    #     dump('hi')

                    #     # Join condition columns
                    #     left = self._column(getattr(entity.table.c, relation.local_key))
                    #     right = self._column(getattr(relation.entity.table.c, relation.foreign_key))

                    #     # Append new Join
                    #     join = Join(
                    #         table=relation.entity.table,
                    #         tablename=str(relation.entity.table.name),
                    #         left=left,
                    #         right=right,
                    #         onclause=left.sacol == right.sacol,
                    #         alias=relation_name,
                    #         method='outerjoin'
                    #     )
                    #     query.joins.append(join)

                    # Add joins only for one-to-one or one-to-many inverse
                    # Outer Join in case some foreign keys are nullable
                    #if field.has_one or field.belongs_to:
                    #if query.joins is None: query.joins = entity.table
                    #query.joins = query.joins.outerjoin(
                    #    right=relation.entity.table,
                    #    onclause=getattr(entity.table.c, relation.local_key) == getattr(relation.entity.table.c, relation.foreign_key)
                    #)

                # Walk down each "include parts" entities
                entity = relation.entity

        # Set query.relations
        query.relations = relations

    def _build_orm_results(self, query: Query, primary: List, secondary: Dict = {}) -> List[E]:
        # No primary results, return empty List
        if not primary: return []

        self.log.nl().header('Relations')
        self.log.dump(query.relations)

        self.log.nl().header('Primary Results')
        self.log.dump(primary)

        self.log.nl().header('Has Many Data')
        self.log.dump(secondary)

        # Deepcopy relations Dict so I can remove relations I have already processed.
        # I process all secondary results first.  This means all left over relations are of
        # the primary results
        relations = deepcopy(query.relations)

        # Dictionary of all secondary converted models
        models = {}

        # Dictionary of all *One models as a cache to deduplicate class instantiation
        singles = {}

        # Full any *One relations method
        def fill_one_relations(rel_name: str, data: List):
            """Fill only the *One relations (One-To-One, One-To-Many)"""
            self.log.nl().header('Filling *One Relations for ' + rel_name)

            # Skip if no data
            if not data:
                #models[rel_name] = {}
                return

            # Determin if data is primary or secondary
            primary = (rel_name == 'primary')

            # Split rel_name into __ parts
            rel_name_parts = rel_name.split('__')

            # Get the actual field, relation and entity for this relation rel_name
            entity = self.entity
            if not primary:
                for rnpart in rel_name_parts:
                    field: Field = entity.modelfield(rnpart)
                    entity = field.relation.fill(field).entity

            #self.log.item('Field: ' + str(field))
            self.log.item('Entity: ' + str(entity))
            self.log.item('Data Keys: ' + str(data[0].keys()))

            # Add a new List to our Dict of models
            models[rel_name] = {}

            # Pk field and column
            pk = entity.pk
            pk_column = entity.mapper(pk).column()
            if not primary: pk_column = rel_name + '__' + pk_column

            # Track completed relations so I can remove from our relations list later
            completed_relations = {}

            # Loop each row of raw data
            i = 0
            for row in data:

                # Because of Many-To-Many we could have the same model multiple times.  But we only want
                # to convert and deal with it once based on unique PK
                if getattr(row, pk_column) in models[rel_name]: continue
                #if models[rel_name][getattr(row, pk_column)]: continue

                # Convert this one row to model (just the main fields, not relations)
                if primary:
                    root_model = entity.mapper(row).model()
                    #root_model = entity.mapper(row).row_to_model()
                else:
                    root_model = entity.mapper(row, rel_name).model()
                    #root_model = entity.mapper(row, rel_name).row_to_model()

                # Get pk value
                pk_value = getattr(root_model, pk)

                # Loop only *One relations that apply to this one "data" model
                relation: Relation
                for relation in relations.values():

                    # Only look at relations that begin with this relation__ and are *One
                    if not primary and rel_name + '__' not in relation.name: continue

                    # Walk up relations and exclude if ANY relation is a *Many starting from rel_name and up
                    if relation.contains_many(query.relations, skip=rel_name_parts): continue

                    # Log output
                    if i == 0: self.log.item('Relation: ' + relation.name + ' - ' + str(relation))

                    # RowProxy results lookup prefix
                    prefix = relation.name

                    # Get all relation fieldnames from relation.name split
                    fieldnames = relation.name.split('__')
                    if not primary:
                        # Skip the first __ parts of rel_name
                        fieldnames = fieldnames[len(rel_name_parts):]
                    if i == 0: self.log.item2('  Fieldnames: ' + ', '.join(fieldnames))

                    # Walk down the root model by fieldnames until you reach the nested
                    # model that has the right field to hold this converted sub model
                    # Remember each relation has the full__nested__name so always start with the
                    # root_model for each relation and work your way down.
                    model = root_model
                    for f in range(0, len(fieldnames) - 1):
                        model = getattr(model, fieldnames[f])

                    # Actual fieldname is always the LAST of the fieldnames, but our model was walked down.
                    fieldname = fieldnames[-1] if fieldnames else relation.name

                    # Walkdown Log
                    if i == 0: self.log.item2('  Model Field: ' + fieldname)
                    if i == 0: self.log.item2('  Field Model: ' + str(model.__class__))

                    # Convert this one rows relation data into a sub_relation model
                    # Only convert each unique *One record just once, or else pull from singles cache
                    # The odd part about this cache is if you include many nested relations one one parent model
                    # And a child model also uses the same child, it too will include all nested relations
                    # Example if you do .include('creator.info', 'owner').  If owner is id=1 and id=1 was already
                    # a creator, that owner will also have the nested INFO filled out, because it pulls from the cache.
                    if relation.entity.tablename not in singles: singles[relation.entity.tablename] = {}
                    sub_model_pk = relation.name + '__' + relation.entity.mapper(relation.entity.pk).column()
                    sub_model_pk_value = getattr(row, sub_model_pk)
                    if sub_model_pk_value is not None and sub_model_pk_value not in singles[relation.entity.tablename]:
                        singles[relation.entity.tablename][sub_model_pk_value] = relation.entity.mapper(row, prefix).model()
                        #singles[relation.entity.tablename][sub_model_pk_value] = relation.entity.mapper(row, prefix).row_to_model()

                    # Get sub_model from singles cache
                    if sub_model_pk_value is not None:
                        sub_model = singles[relation.entity.tablename][sub_model_pk_value]
                        #sub_model = relation.entity.mapper(row, prefix).model(False) # No cache version

                        # Add this converted sub_model to the walked down parent model
                        setattr(model, fieldname, sub_model)

                    # Mark relation as complete so I can delete from relations Dict later
                    completed_relations[relation.name] = 1

                # All *One relations have been converted and merged
                # Add this fully converted (including nested *One relations) model to List of models
                models[rel_name][pk_value] = root_model
                i += 1

            # Delete all completed relations from our relation deepcopy.  We will not need them again
            for completed_relation in completed_relations.keys():
                del relations[completed_relation]

        # Fill in all *One relations for all secondary results first.
        # as each relation is merged it will be removed from our local relations deepcopy.
        # All relations left will be those on the main results data.
        for rel_name, rel_data in secondary.items():
            fill_one_relations(rel_name, rel_data)
        #fill_one_relations(secondary['comments'], 'comments')

        # Fill in all *One relations for the primary results.
        fill_one_relations('primary', primary)

        # All relations left should be of *Many either for the primary results
        # or for any of the secondary results
        self.log.nl().header('Leftover Relations are *Many')
        self.log.dump(relations)

        # All records in models Dict are *Many and the main Primary dataset
        # All remaining relations are the *Many which should match the models Dict key
        # Looping the *Many relations in REVERSE gives us the deepest relations first which is critical
        self.log.nl().header('Combining Recursive *Many Models')

        relation: Relation
        for relation in reversed(relations.values()):

            # Relation name parts
            relation_parts = relation.name.split('__')
            field = relation_parts[-1]

            # If models does not contain this relation, skip the merge
            if relation.name not in models: continue

            # Get parent and child Dict of models
            children_name = relation.name
            children = models[relation.name]
            parents_name = 'primary'
            if len(relation_parts) > 1: parents_name = '__'.join(relation_parts[:-1])
            if parents_name in models:
                # Parent is a *Many so grab from models
                parents = models[parents_name]
            else:
                # Parent is a *One, so grap from singles cache
                parents = singles[query.relations.get(parents_name).entity.tablename]

            self.log.item('Combining child: ' + children_name + ' into parent: ' + parents_name)

            # Determine if child *Many results should be displayed as a Dict or List
            dict_key = getvalue(relation, 'dict_key')
            dict_value = getvalue(relation, 'dict_value')
            list_value = getvalue(relation, 'list_value')

            # Loop parents so we can at least set each child to empty [] instead of None.  We always want [] instead of None for empty children
            for parent in parents.values():
                # Set empty [] or {}
                if dict_key:
                    setattr(parent, field, {})
                else:
                    setattr(parent, field, [])

            # Merge in Many-To-Many by using the original RowProxy result which contains
            # The pivot tables joining column (left_key)
            if type(relation) == BelongsToMany or type(relation) == MorphToMany:
                #left_key = relation.name + '__' + relation.left_key
                left_key = relation.name + '__' + relation.entity.mapper(relation.left_key).field()
                right_key = relation.name + '__' + relation.entity.mapper(relation.entity.pk).column()

                # QUESTION, what is the differente from children.values() vs
                # secondary?  Look at the ELSE below that for child in children.values()
                # but this many* uses secondary?
                # Can I combine all relations into one large loop?
                # Because I am doing identical work in the dict_key stuff

                # ALSO all of this dict_key code may not work anyway
                # I bet the API will not know how to handle input and complain?
                # I may have to handle specially in the ModelRouter

                # Loop raw RowProxy to find proper pivot keys
                for row in secondary[relation.name]:
                    left_id = getattr(row, left_key)
                    right_id = getattr(row, right_key)

                    # Get parent value, the value of the main table
                    parent = parents[left_id]

                    # Get child value. The value of the many record
                    child = children[right_id]

                    # Set None field to empty List
                    if getattr(parent, field) is None:
                        setattr(parent, field, [])

                    # Add each *Many model as a Dict
                    if dict_key:
                        if dict_value:
                            if type(dict_value) == list:
                                value = {key:getattr(child, key) for key in dict_value}
                            else:
                                value = getattr(child, dict_value)
                        else:
                            value = child.dict()
                        getattr(parent, field)[getattr(child, dict_key)] = value


                    # Add each *Many model as a List of a single value
                    elif list_value:
                        getattr(parent, field).append(getattr(child, list_value))

                    # Add each *Many as a List of the actual Models
                    else:
                        # Append to list using deepcopy
                        getattr(parent, field).append(
                            # We must deep copy the record becuase we dedup the *Many
                            # but they could be used multiple times
                            # ?? Hum maybe not, good if you change one it changes them all
                            # But all the *One will NOT be like this I don't believe, have to test
                            #deepcopy(child[right_id])
                            child
                        )

            else:

                for child in children.values():
                    parent_pk_value = getattr(child, relation.entity.mapper(relation.foreign_key).field())
                    if parent_pk_value not in parents: continue;

                    parent = parents[parent_pk_value]
                    field = relation_parts[-1]

                    # # Set None field to empty list
                    # if getattr(parent, field) is None:
                    #     setattr(parent, field, [])

                    # DUPLCATING work here, see above for nearly exact same thing
                    # need to optimize this code

                    # Add each *Many model as a Dict
                    if dict_key:
                        if dict_value:
                            if type(dict_value) == list:
                                # Dict value is a list.  Create a dictionary from the lists keys
                                value = {key:getattr(child, key) for key in dict_value}
                            else:
                                # Dict value is a string, use just that fields value
                                value = getattr(child, dict_value)
                        else:
                            # No dict value set, but there is a dict_key, so we want a dict.  Use the entire record as a dict
                            value = child.dict()
                        getattr(parent, field)[getattr(child, dict_key)] = value
                        #setattr(parent, field, 'x')

                    # Add each *Many model as a List of a single value
                    elif list_value:
                        getattr(parent, field).append(getattr(child, list_value))

                    # Add each *Many as a List of the actual Models
                    else:
                        getattr(parent, field).append(child)


        self.log.nl().header('Singles Cache')
        self.log.dump(singles)

        self.log.nl().header('Secondary *Many Models')
        self.log.dump(models)

        # These models are already a Dict keyed by PK
        # Return existing primary model if user wanted keyby id
        if query.keyed_by == self.entity.pk:
            return models['primary']

        # Key results by another column
        if query.keyed_by:
            keyed_entities = {}
            for entity in models['primary'].values():
                keyed_entities[getattr(entity, self.query.keyed_by)] = entity

            # Return Dictionary of Entities
            return keyed_entities

        # No keyby, convert primary models to a List
        return [x for x in models['primary'].values()]

    def _connection(self):
        return self.entity.connection

    def _pk(self):
        return self.entity.pk

    def _column_from_string(self, dotname: str, query: Query) -> Tuple:
        if '.' in dotname:
            parts = dotname.split('.')
            relation = query.relations.get('__'.join(parts[:-1]))
            table = self._get_join_table(query, alias=relation.name)  # Get table from join alias since its a relation
            field = parts[-1]
            name = self.entity.mapper(field).column()
        else:
            name = self.entity.mapper(dotname).column()
            table = query.table

        tablename = str(table.name)
        column = table.columns.get(name)
        return (table, tablename, column, name, self._connection())

    def _get_join_table(self, query: Query, alias: str):
        """Get the join table for this table alias"""
        for join in query.joins:
            if join.alias == alias:
                return join.table

        # No, found that we ALWAYS want by alias to prevent grabbing the wrong duplicate named table
        # Aliases are always unique
        # if tablename is not None:
        #     for join in query.joins:
        #         if join.tablename == tablename:
        #             return join.table

        #     # Not found by name, check alias
        #     for join in query.joins:
        #         if join.alias == tablename:
        #             return join.table

    def _get_tablename(self, table: Union[sa.Table, sa.sql.selectable.Alias]):
        """Get the REAL tablename even if the table is an Alias"""
        if type(table) == sa.sql.selectable.Alias:
            return table.original.name
        else:
            return table.name

# IoC Class Instance
#_OrmQueryBuilderIoc: _OrmQueryBuilder = uvicore.ioc.make('OrmQueryBuilder', _OrmQueryBuilder)

# Actual Usable Model Class Derived from IoC Inheritence
#class OrmQueryBuilder(Generic[B, E], _OrmQueryBuilderIoc[B, E], BuilderInterface[B, E]):
    #pass



# type: ignore
from .application import Application
from .builder import DbQueryBuilder, OrmQueryBuilder, QueryBuilder
from .cache import Cache
from .config import Config
from .connection import Connection
from .database import Database
from .dispatcher import Dispatcher
#from .event import Event
from .field import Field
from .ioc import Binding, Ioc
from .job_dispatcher import JobDispatcher
from .job import Job
from .logger import Logger
from .mapper import Mapper
from .model import Model
from .package import Package
from .provider import Provider
from .relation import Relation
#from .router import ApiRouter, WebRouter
#from .routes import Routes
from .router import (ApiRoute, ApiRouter, ModelRouter, Router, Routes,
                     WebRoute, WebRouter)
from .server import Server
from .template import Template
from .user_info import UserInfo
from .user_provider import UserProvider
from .authenticator import Authenticator
from .email import Email
from .auto_api import AutoApi

# The contracts package uses a __init__.py because users will import these
# methods from their own apps and we want a nicer import which looks
# like - from uvicore.contracts import Model



import re
import copy
from uvicore.support.dumper import dump, dd
from collections import OrderedDict as _OrderedDict
from prettyprinter import pretty_call, register_pretty


# Builtin Dict Methods
# clear
# copy
# popitem
# setdefault
# update
# fromkeys


class _SuperDict:
    """Advanced python dictionary and ordered dictionary replacement"""
    # A refactor of https://github.com/mewwts/addict (MIT)

    def __init__(__self, *args, **kwargs):
        object.__setattr__(__self, '__parent', kwargs.pop('__parent', None))
        object.__setattr__(__self, '__key', kwargs.pop('__key', None))
        object.__setattr__(__self, '__frozen', False)
        for arg in args:
            if not arg: continue

            if isinstance(arg, dict):
                for key, val in arg.items():
                    __self[key] = __self._hook(val)
            elif isinstance(arg, tuple) and (not isinstance(arg[0], tuple)):
                __self[arg[0]] = __self._hook(arg[1])
            else:
                for key, val in iter(arg):
                    __self[key] = __self._hook(val)

        for key, val in kwargs.items():
            __self[key] = __self._hook(val)


    def __call__(self, dotkey: str = None):
        return self.dotget(dotkey)

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __setattr__(self, name, value):
        """Set new attribute and detect if it is a reserved keyword if the dict class"""
        if hasattr(self.__class__, name):
            msg = "'Dict' object attribute '{}' is read-only"
            raise AttributeError(msg.format(name))
        self[name] = value

    def __delattr__(self, name):
        del self[name]

    def __setitem__(self, name, value):
        # Check if dict is frozen (read only)
        frozen = (hasattr(self, '__frozen') and object.__getattribute__(self, '__frozen'))
        if frozen and name not in super(_SuperDict, self).keys():
            msg = 'Dictionary is frozen / readonly.'
            raise KeyError(msg)

        super(_SuperDict, self).__setitem__(name, value)
        try:
            p = object.__getattribute__(self, '__parent')
            key = object.__getattribute__(self, '__key')
        except AttributeError:
            p = None
            key = None
        if p is not None:
            p[key] = self
            object.__delattr__(self, '__parent')
            object.__delattr__(self, '__key')

    def __add__(self, other):
        if not self.keys(): return other

        self_type = type(self).__name__
        other_type = type(other).__name__
        msg = "Unsupported operand type(s) for +: '{}' and '{}'"
        raise TypeError(msg.format(self_type, other_type))

    def __missing__(self, name):
        # If missing, create a new Dict, unless frozen
        #frozen = (hasattr(self, '__frozen') and object.__getattribute__(self, '__frozen'))
        frozen = object.__getattribute__(self, '__frozen')
        if frozen:
           raise KeyError(name)
        # Don't use self to make a new blank key as it could be OrderedDict
        # We always want a missing value to be an actual SuperDict
        #return self.__class__(__parent=self, __key=name)
        return Dict(__parent=self, __key=name)

    def __getnewargs__(self):
        return tuple(self.items())

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self.update(state)

    def __or__(self, other):
        if not isinstance(other, (_SuperDict, dict)):
            return NotImplemented
        new = self.__class__(self)
        new.update(other)
        return new

    def __ror__(self, other):
        if not isinstance(other, (_SuperDict, dict)):
            return NotImplemented
        new = self.__class__(other)
        new.update(self)
        return new

    def __ior__(self, other):
        self.update(other)
        return self

    def __deepcopy__(self, memo):
        other = self.__class__()
        memo[id(self)] = other
        for key, value in self.items():
            other[copy.deepcopy(key, memo)] = copy.deepcopy(value, memo)
        return other

    @classmethod
    def _hook(cls, item):
        if isinstance(item, _SuperDict):
            # Already a SuperDict (Dict or OrderedDict) don't reconvert
            return item
        elif isinstance(item, dict):
            # A standard builtin dict, convert to SuperDict
            # Do NOT however use cls(item) as that will convert to the same type
            # this one is.  In the case of OrderedDict all sub dicts will become
            # OrderedDicts, which we don't want.  Only explicity defined OrderedDicts
            # should be kept intact.  Any sub dict under that should be an our Dict
            #return cls(item)
            return Dict(item)
        elif isinstance(item, (list, tuple)):
            # A List or tuple
            return type(item)(cls._hook(elem) for elem in item)
        return item

    def copy(self):
        """Make a byvalue copy of entire dict

        This overrides the builtin dict method."""
        return copy.copy(self)

    def deepcopy(self):
        """Make a byvalue copy of entire dict"""
        return copy.deepcopy(self)

    def clone(self):
        """Alias of deepcopy"""
        return self.deepcopy()

    def update(self, *args, **kwargs):
        """Deep merge one or more dictionaries with self.  Overwrites values that exist.

        This overrides the builtin dict method."""
        other = {}
        if args:
            if len(args) > 1: raise TypeError()
            # Add in arguments (assumed a dict), convert to SuperDict
            other.update(self.__class__(args[0]))
        other.update(kwargs)
        for k, v in other.items():
            if ((k not in self) or (not isinstance(self[k], dict)) or (not isinstance(v, dict))):
                # Add in any NON dictionary items that do not exist
                self[k] = v
            else:
                # Add in dictionary items, convert to SuperDict
                self[k].update(self.__class__(v))
        return self

    def merge(self, *args, **kwargs):
        """Deep merge one or more dictionaries with self.  Overwrites values that exist."""
        self.update(*args, **kwargs)
        return self

    def defaults(self, *args, **kwargs):
        """Provide defaults Dict to existing self and add only if not exists.
        Essentially a reverse merge that does NOT overwrite values if exist"""
        defaults = self.__class__(*args, **kwargs)
        defaults.merge(self)
        self.merge(defaults)
        return self

    def setdefault(self, key: str, default=None):
        """Set a single key value only if not exists.

        This overrides the builtin dict method."""
        # if key in self:
        #     return self[key]
        # else:
        #     self[key] = default
        #     return default

        if key not in self:
            self[key] = default
        return self

    def to_dict(self):
        """Convert SuperDict into regular builtin dict"""
        base = {}
        for key, value in self.items():
            if isinstance(value, type(self)):
                base[key] = value.to_dict()
            elif isinstance(value, (list, tuple)):
                base[key] = type(value)(
                    item.to_dict() if isinstance(item, type(self)) else
                    item for item in value)
            else:
                base[key] = value
        return base

    def dict(self):
        """Convert SuperDict into regular builtin dict"""
        # Why to_dict() and dict().  Pydantic has a dict(), so I want to mimic
        # so all calls to dict() on SuperDict or Pydanttic models work the same
        # Mostly for uvicore orm usage.
        return self.to_dict()

    def freeze(self, freeze=True):
        """Set entire SuperDict to read only"""
        object.__setattr__(self, '__frozen', freeze)
        for key, val in self.items():
            if isinstance(val, _SuperDict):
                val.freeze(freeze)

    def unfreeze(self):
        """Set entire SuperDict to writeable"""
        self.freeze(freeze=False)

    def dotget(self, dotkey: str = None, _recursive_config = None):
        """Get values using string dot notation.  Surround with [] to include those . as the key"""
        # Recursive for dot notation
        if not dotkey: return self
        if _recursive_config is None: _recursive_config = self

        try:

            # Replace any , with ^ inside [] temporarily
            if "[" in dotkey:
                def x(m): return m[0].replace('.', '^')[1:-1]
                dotkey = re.sub('(\[.*?\])', x, dotkey)

            if "." in dotkey:
                key, rest = dotkey.split(".", 1)
                key = key.replace('^', '.')
                if key not in _recursive_config:
                    _recursive_config[key] = self.__class__()
                return self.dotget(rest, _recursive_config[key])
            else:
                dotkey = dotkey.replace('^', '.')
                if dotkey in _recursive_config:
                    return _recursive_config[dotkey]
                else:
                    return self.__missing__(dotkey)
        except:
            return self.__missing__(dotkey)

    def dotset(self, dotkey: str, value, _recursive_config=None):
        """Set values using string dot notation.  Surround with [] to include those . as the key."""
        # Recursive for dot notation
        # Remember objects are byRef, so changing config also changes self.items
        if _recursive_config is None: _recursive_config = self

        # Replace any , with ^ inside [] temporarily
        if "[" in dotkey:
            def x(m): return m[0].replace('.', '^')[1:-1]
            dotkey = re.sub('(\[.*?\])', x, dotkey)

        if "." in dotkey:
            key, rest = dotkey.split(".", 1)
            key = key.replace('^', '.')
            if key not in _recursive_config:
                _recursive_config[key] = self.__class__()
            return self.dotset(rest, value, _recursive_config[key])
        else:
            # Replace ^ back with . for keys inside []
            dotkey = dotkey.replace('^', '.')
            if type(value) == dict:
                _recursive_config[dotkey] = self.__class__(value)
            else:
                _recursive_config[dotkey] = value
            return value


class Dict(_SuperDict, dict):
    """Dictionary that you can access like a class using dot notation attributes"""
    # def __init__(__self, *args, **kwargs):
    #     @register_pretty(__self.__class__)
    #     def pretty_entity(value, ctx):
    #         """Custom pretty printer for my SuperDict"""
    #         # This printer removes the class name uvicore.types.Dict and makes it print
    #         # with a regular {.  This really cleans up the output!

    #         #return pretty_call(ctx, cls, **value.__dict__)
    #         return pretty_call(ctx, '', value.to_dict())

    #     super().__init__(*args, **kwargs)
    pass



class OrderedDict(_SuperDict, _OrderedDict):
    """Ordered Dictionary that you can access like a class using dot notation attributes"""

    # def __init__(__self, *args, **kwargs):

    #     @register_pretty(__self.__class__)
    #     def pretty_entity(value, ctx):
    #         """Custom pretty printer for my SuperDict"""
    #         # This printer removes the class name uvicore.types.Dict and makes it print
    #         # with a regular {.  This really cleans up the output!

    #         #return pretty_call(ctx, cls, **value.__dict__)
    #         return pretty_call(ctx, __self.__class__, value.to_dict())

    #     super().__init__(*args, **kwargs)
    pass


@register_pretty(Dict)
def pretty_entity(value, ctx):
    """Custom pretty printer for my SuperDict"""
    # This printer removes the class name uvicore.types.Dict and makes it print
    # with a regular {.  This really cleans up the output!

    #return pretty_call(ctx, 'Dict', **value)
    #return pretty_call(ctx, 'Dict', value.to_dict())  # Does show nested dicts as SuperDicts
    return pretty_call(ctx, 'Dict', dict(**value))



@register_pretty(OrderedDict)
def pretty_entity(value, ctx):
    """Custom pretty printer for my SuperDict"""
    # This printer removes the class name uvicore.types.Dict and makes it print
    # with a regular {.  This really cleans up the output!

    #return pretty_call(ctx, 'OrderedDict', **value)
    #return pretty_call(ctx, 'OrderedDict', value.to_dict())  # Does show nested dicts as SuperDicts
    return pretty_call(ctx, 'OrderedDict', dict(**value))





from abc import ABC, abstractmethod
from typing import Dict, List, Any, Optional


class Template(ABC):

    @property
    @abstractmethod
    def env(self) -> Any: pass

    @property
    @abstractmethod
    def paths(self) -> List[str]: pass

    @property
    @abstractmethod
    def context_functions(self) -> Dict: pass

    @property
    @abstractmethod
    def context_filters(self) -> Dict: pass

    @property
    @abstractmethod
    def filters(self) -> Dict: pass

    @property
    @abstractmethod
    def tests(self) -> Dict: pass

    def render(self, template_name: str, data: Optional[Dict] = {}) -> str:
        """Render a template as string (for CLI usage, not a Web response)"""
        pass

    def render_web_response(self, name: str, context: dict, status_code: int = 200, headers: dict = None, media_type: str = None, background = None):
        """Render a template as Web Response"""
        pass



import hashlib
from typing import Any

def md5(item: Any):
    """Create an md5 hash of the given item"""
    return hashlib.md5((str(item)).encode('utf-8')).hexdigest()

def sha1(item: Any):
    return hashlib.sha1((str(item)).encode('utf-8')).hexdigest()

def sha256(item: Any):
    return hashlib.sha256((str(item)).encode('utf-8')).hexdigest()

def sha512(item: Any):
    return hashlib.sha512((str(item)).encode('utf-8')).hexdigest()



import uvicore
from argon2 import Type
from uvicore.typing import Dict
from argon2 import PasswordHasher
from uvicore.support.dumper import dump, dd
from argon2.exceptions import VerifyMismatchError


def create(password: str, *, type: str = None, hash_len: int = None, salt_len: int = None, encoding: str = None, time_cost: int = None, memory_cost: int = None, parallelism: int = None):
    """Create a new hashed password using config driven Argon2 default or parameter overrides"""
    config = uvicore.config('uvicore.auth.hasher')
    if type is None: type = config.type
    if hash_len is None: hash_len = config.hash_len
    if salt_len is None: salt_len = config.salt_len
    if encoding is None: encoding = config.encoding
    if time_cost is None: time_cost = config.time_cost
    if memory_cost is None: memory_cost = config.memory_cost
    if parallelism is None: parallelism = config.parallelism

    # New Argon2 Password Hasher
    ph = PasswordHasher(type=Type[type], hash_len=hash_len, salt_len=salt_len, encoding=encoding, time_cost=time_cost, memory_cost=memory_cost, parallelism=parallelism)
    return ph.hash(password)

def verify(password: str, hash: str) -> bool:
    if password is None or hash is None: return False
    ph = PasswordHasher()
    verified = False
    try:
        verified = ph.verify(hash, password)
    except VerifyMismatchError:
        pass
    return verified



import uvicore
from uvicore.typing import Dict
from uvicore.events import Handler
from uvicore.support.module import load
from uvicore.support.dumper import dump, dd
from uvicore.console import group as cli_group
from uvicore.foundation.events.app import Booted as OnAppBooted


class Console(Handler):

    def __call__(self, event: OnAppBooted):
        """Bootstrap the Console after the Application is Booted"""

        #from uvicore.console.console import cli
        cli = uvicore.ioc.make('uvicore.console.console.cli')

        # Deep merge all command groups, this allows simple command extension!
        groups = Dict()
        for package in uvicore.app.packages.values():
            if not 'console' in package: continue

            # Only register commands if package registration is allowed
            # AND if running from the console...OR if the command group is
            # 'http'.  We always want the http group regardless of cli or
            # http entrypoints.
            register = package.registers.commands
            is_console = uvicore.app.is_console
            for key, group in package.console['groups'].items():
                if (register and is_console) or key == 'http':
                    groups.merge({key: group})

        # Register each group and each groups commands
        click_groups = {}
        for key, group in groups.items():
            # Split key to find parent/child groups
            parts = key.split(':')
            parent = ':'.join(parts[0: -1])
            name = parts[-1]

            # Create a new click group if not exists
            @cli_group(help=group['help'])
            def click_group():
                pass
            click_groups[key] = click_group

            # Get click group based on key (this is how we handle sub groups)
            #click_group = click_groups[key]

            # Add all commands into this click_group
            for command_name, command_module in group.commands.items():
                # Dynamically import the commands module
                module = load(command_module).object
                click_group.add_command(module, command_name)

            # Add group to console
            if len(parts) == 1:
                # Root level group
                cli.add_command(click_group, name)
            else:
                # Sub group
                click_groups[parent].add_command(click_group, name)



from typing import Callable, List

from fastapi import FastAPI as _FastAPI
from starlette.applications import Starlette as _Starlette
from starlette.types import ASGIApp

import uvicore
from uvicore.contracts import Server as ServerInterface
from uvicore.support.dumper import dd, dump

# This is our actual FastAPI Server
# Can only have one server, can't be both FastAPI and/or Starlette
# The FastAPI server inherits Starlette server but extends it with
# extra options for openapi, so we must use that as our base http app

# In classic FastAIP example, this is same as
# app = FastAPI(), but I call it app.http

# Bound from service provider instead of decorator


# @uvicore.service('uvicore.http.server._Server',
#     aliases=['http', 'HTTP'],
#     singleton=True,
#     kwargs={
#         'debug': uvicore.config('app.debug'),
#         'title': uvicore.config('app.openapi.title'),
#         'version': uvicore.app.version,
#         'openapi_url': uvicore.config('app.openapi.url'),
#         'docs_url': uvicore.config('app.openapi.docs_url'),
#         'redoc_url': uvicore.config('app.openapi.redoc_url'),
#     },
# )
# class _Server(_FastAPI):
#     @property
#     def server(self) -> _FastAPI:
#         return self


    # def include_router(self, router, *, prefix: str = '', tags: List[str] = None) -> None:
    #     # _server is a FastAPI application.  If you look at FastAPIs include_router
    #     # it automatically detects if the the actual Router is FastAPI or Starlette
    #     # and handles the including of the routers routes differently!

    #     for route in router.routes:
    #         self.add_route(
    #             path=route.pathroute)

        # self._server.include_router(
        #     router=router,
        #     prefix=prefix,
        #     tags=tags,
        # )


        # self,
        # path: str,
        # route: typing.Callable,
        # methods: typing.List[str] = None,
        # name: str = None,
        # include_in_schema: bool = True,






# Example of abstratsion
class ServerXXXX(ServerInterface):
    """HTTP Server private class.

    Do not import from this location.
    Use the uvicore.app.http singleton global instead."""

    @property
    def server(self) -> _FastAPI:
        return self._server

    def __init__(self, debug: bool, title: str, version: str, openapi_url: str, docs_url: str, redoc_url: str):

        # Fireup FastAPI HTTP Core
        self._server = _FastAPI(
            debug=debug,
            title=title,
            version=version,
            openapi_url=openapi_url,
            docs_url=docs_url,
            redoc_url=redoc_url,
        )

    def include_router(self, router, *, prefix: str = '', tags: List[str] = None) -> None:
        # _server is a FastAPI application.  If you look at FastAPIs include_router
        # it automatically detects if the the actual Router is FastAPI or Starlette
        # and handles the including of the routers routes differently!
        self._server.include_router(
            router=router,
            prefix=prefix,
            tags=tags,
        )

    def mount(self, path: str, app: ASGIApp, name: str = None) -> None:
        self._server.mount(path=path, app=app, name=name)

    def on_event(self, event_type: str) -> Callable:
        return self._server.on_event(event_type)


# IoC Class Instance
# No because not to be used by the public
#Server: ServerInterface = uvicore.ioc.make('Http')

# Public API for import * and doc gens
#__all__ = ['_Server']



import uvicore
from uvicore.orm.fields import Field
from uvicore.support.dumper import dd, dump
from prettyprinter import pretty_call, register_pretty
from pydantic.fields import FieldInfo as PydanticFieldInfo
from pydantic.main import ModelMetaclass as PydanticMetaclass
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

import sqlalchemy as sa
from sqlalchemy.sql import ClauseElement

# try:
# except ImportError:  # pragma: nocover
#     pass
#     #sa = None  # type: ignore
#     #ClauseElement = None  # type: ignore


# Think of this metaclass as all the STATIC methods similar to @classmethod
# but different in that they become INVISIBLE to the instance of the class
# which is ideal so pydantic doesn't see these and clobber any actual 'fields'
# that have the same name.  This means I can have a field called 'email' and a
# true static class method called `email()` and pydantic won't complain.  I can use
# the static via User.email() and the instatnce from user.email.
# See https://stackoverflow.com/questions/59341761/what-are-the-differences-between-a-classmethod-and-a-metaclass-method
# for a detailed difference between using @classmethod for statics vs a metaclass


@uvicore.service()
class ModelMetaclass(PydanticMetaclass):

    # Testing of duplicate field
    # Only works if EACH model sets a metaclass=ModelMetaclass
    # You cannot set the metaclass at the Module.py level
    # def email(entity):
    #     return 'email on meta'

    # def slug(entity):
    #     return 'slug on meta'


    # Remember all metaclass method ARE @classmethods
    # Remember any method here DO NOT clash with pydantic field names
    # Remember any method here do NOT show up in code intellisense because its a metaclass
    #   If you want it in code intellisense, it must be in model.py and ModelInterface
    #   but then it WILL clash with any pydantic fields of the same name.
    # Basically anything here should be used privately from inside the ORM.
    #   If any method is to be public (either as a @classmethod Post. or instance method post.
    #   it should go in model.py and be in the ModelInterface



    ############################################################################
    ## These query builder passthroughs are for convenience only.  Best to use
    ## MyModel.query().get()... instead (the query() is on the Model.py not
    ## this metaclass).  Why use query()?  Becuase metaclasses are not currently
    ## supported in VSCode for autocomplete code intellisense but parent classes
    ## are.  So by using query() you get autocomplete on the full query builder!
    ############################################################################
    # async def get(entity) -> List[Any]:
    #     """Query builder passthrough"""
    #     return await OrmQueryBuilder(entity).get()

    # async def find(entity, id: Any) -> Any:
    #     """Query builder passthrough"""
    #     return await OrmQueryBuilder(entity).find(id)

    # def where(entity, column: Union[str, List[Tuple]], operator: str = None, value: Any = None):
    #     """Query builder passthrough"""
    #     return OrmQueryBuilder(entity).where(column, operator, value)

    # def or_where(entity, wheres: List):
    #     """Query builder passthrough"""
    #     return OrmQueryBuilder(entity).or_where(wheres)

    # def include(entity, *args):
    #     """Query builder passthrough"""
    #     return OrmQueryBuilder(entity).include(*args)
    ############################################################################
    ############################################################################


    @property
    def pk(entity) -> str:
        """Get the entities primary key"""
        for field in entity.modelfields.values():
            if field.primary: return field.name

    @property
    def connection(entity) -> str:
        """Helper for entity connection string"""
        return entity.__connection__

    @property
    def tablename(entity) -> str:
        """Helper for entity tablename string"""
        return entity.__tablename__

    @property
    def table(entity) -> sa.Table:
        """Helper for entity SQLAlchemy table"""
        return entity.__table__

    @property
    def modelfields(entity) -> Dict[str, Field]:
        """Helper for original uvicore model fields (not pydantic __fields__)"""
        return entity.__modelfields__

    @property
    def modelname(entity) -> str:
        return entity.__name__

    @property
    def modelfqn(entity) -> str:
        module = entity.__module__
        if module is None or module == str.__module__:
            return entity.__name__  # Avoid reporting __builtin__
        else:
            return module + '.' + entity.__name__

    def modelfield(entity, fieldname: str) -> Field:
        field = entity.modelfields.get(fieldname)
        if not field: raise Exception("Field {} not found in model {}".format(fieldname, entity.modelfqn))
        return field

    async def execute(entity, query: Union[ClauseElement, str], values: Union[List, Dict] = None) -> Any:
        """Database execute in the context of this entities connection"""
        return await uvicore.db.execute(query=query, values=values, connection=entity.__connection__)

    async def fetchone(entity, query: Union[ClauseElement, str], values: Dict = None) -> Optional[Mapping]:
        """Database fetchone in the context of this entities connection"""
        return await uvicore.db.fetchone(query=query, connection=entity.__connection__)

    async def fetchall(entity, query: Union[ClauseElement, str], values: Dict = None) -> List[Mapping]:
        """Database fetchall in the context of this entities connection"""
        return await uvicore.db.fetchall(query=query, connection=entity.__connection__)

    # def to_model(entity, row, prefix: str = None) -> Any:
    #     """Convert a row of table data into a model"""
    #     fields = {}
    #     for field in entity.modelfields.values():
    #         if not field.column: continue
    #         column = field.column
    #         if prefix: column = prefix + '.' + column
    #         if hasattr(row, column):
    #             fields[field.name] = getattr(row, column)
    #     return entity(**fields)

    # def to_column(entity, fieldname: str):
    #     """Convert a model field name into a table column name"""
    #     field = entity.modelfields.get(fieldname)
    #     if field: return field.column
    #     return fieldname

    def selectable_columns(entity, table: sa.Table = None, *, show_writeonly: Union[bool, List] = False) -> List[sa.Column]:
        """Get all SQLA columns that are selectable

        Why not just use the table to get all columns?  Because a table
        may have far more columns than the actual model.  So we use the model
        to infer a list of actual SQLA columns (excluding write_only fields)
        """
        if table is None: table = entity.table

        # Not all models require tables (databaseless models)
        if table is None: return []

        all_columns = table.columns
        columns: List[sa.Column] = []
        for (field_name, field) in entity.modelfields.items():
            # Exclude None columns (which are relations) and write_only columns which cannot be viewed
            if field.column:
                show = False
                if not field.write_only: show = True
                if type(show_writeonly) == bool:
                    # show_writeonly is a bool, meaning show all writeonly fields
                    if field.write_only and show_writeonly == True: show = True
                else:
                    # show_writeonly is a list of fields to allow
                    if field.write_only and field.column in show_writeonly: show = True

                if show:
                    columns.append(getattr(all_columns, field.column))
        return columns

    def info(entity) -> Dict[str, Any]:
        fields = {}
        for (field_name, field) in entity.modelfields.items():
            fields[field_name] = field
        return {
            'connection': uvicore.db.connection(entity.__connection__),
            'tablename': entity.tablename,
            'table': entity.table,
            'fields': fields,
        }

    def __new__(mcls: type, name: str, bases: Tuple[type, ...], namespace: Dict[str, Any], **kwargs) -> type:
        # mcls is this ModelMetaclass itself
        # name is the string name of the child class, in this case '_Model' from model.py
        # bases is a tuple of parent used in the child (_Model) class, in this case (uvicore.contracts.model.Model, pydantic.main.BaseModel) it does not include this metaclass
        # namespace is the child _Model classes original __dict__ Dictionary

        # Define our custom attributes
        __connection__ = None
        __tablename__ = None
        __table__ = None
        __tableclass__ = None
        __callbacks__ = {}

        # Pull out all model properties of type Field() and store in __modelfields__ property
        # Then replace the original properties from Field() to Pydantics FieldInfo(), converting some Field() arguments
        # into 'x-tra' arguments for FieldInfo()
        # In the end, __modelfields__ are the original uvicore Field() that you defined in the model
        # and the original model fields are replaced with pydantics FieldInfo().  Later when I call super().__new__
        # pydantic further converts by MOVING the actual class properties into __fields__ and changing from FieldInfo to ModelField
        __modelfields__: Dict[str, Field] = {}
        for field_name, field in namespace.items():
            if field_name[0] != '_' and type(field) == Field:
                # Pull out uvicore model Field() into own __modelfields__ dict
                field.name = field_name
                __modelfields__[field_name] = field

                # Derive extra field information for pydantics FieldInfo OpenAPI schema generation
                #dump(getattr(namespace, field_name))

                # Convert uvicore model field into pydantic FieldInfo
                # Be careful what you pipe into PydanticFieldInfo() as these show up in the OpenAPI schema
                # from FastAPI 0.65.0+ because in their openapi/models.py they set class Config extra='allow'
                # by default.  Which means any item you put in PydanticFieldInfo that is not a keyword on that class
                # goes into extra *kwargs and extra now shows up in openapi.json.  The problem with that is valid
                # OpenAPI only allows certain keywords, see https://swagger.io/docs/specification/data-models/keywords/
                # In fact, the 'properties' I was using for a catch-all JSON blog is totally invalid.  Its not a ignored blog
                # its a nested schame field properties.  The only valid way to add extra ignored fields to a schema is to
                # prefix them with x-  see https://swagger.io/specification/#specification-extensions

                # For some dumb reason I was also piping EVERY thing into PydanticFieldInfo, even callback, evaluate
                # and relation.  Which now that extra=allow those were all trying to be parsed by FastAPI json_encoder
                # which totally broke everything.  So I need to be more strategic which uvicore Field() items I pump into
                # Pydantic FieldInfo().  I added a fields_passed_to_pydantic_FieldInfo in my Field class that contains a
                # List of valid properties to pass into PydanticFieldInfo()

                field_info_kwargs = {}
                #dump(field.__annotations__)
                for slot in field.__annotations__.keys():
                    value = getattr(field, slot)
                    if value is None: continue;

                    # Only pipe certain uvicore Field properties into pydantics FieldInfo or you get invalid OpenAPI spec
                    if slot in Field.__valid_oepnapi_keywords__:
                        arg = slot

                        # Convert some understores to camelCase for OpenAPI keyword compatibility
                        if slot == 'read_only': arg = 'readOnly'
                        if slot == 'write_only': arg = 'writeOnly'
                        if slot == 'min_length': arg = 'minLength'
                        if slot == 'max_length': arg = 'maxLength'
                        field_info_kwargs[arg] = value

                    elif slot in Field.__convert_to_extensions__:
                        # Convert these Field() arguments to the x-tra Dict
                        # See https://swagger.io/specification/#specification-extensions
                        if 'x-tra' not in field_info_kwargs: field_info_kwargs['x-tra'] = {}
                        if slot == 'properties':
                            field_info_kwargs['x-tra'] = {**field_info_kwargs['x-tra'], **getattr(field, slot)}
                        else:
                            field_info_kwargs['x-tra'][slot] = getattr(field, slot)

                #dump(field_info_kwargs)
                namespace[field_name] = PydanticFieldInfo(**field_info_kwargs)
                #dump(namespace[field_name])

        #dump(namespace)

        # If we extend and overwrite our own models, then some information
        # will be buried in the bases tuple.  Loop each base and
        # pluck out these critical fields (for modelfields, APPEND them to allow extension)
        for base in bases:
            if hasattr(base, '__modelfields__'):
                # Notice we are APPEND modelfields to allow model extension
                __modelfields__ = {**base.__modelfields__, **__modelfields__}

            # I am only setting these if not already set.  This allows a higher base
            # like the parent to override and win over the base children
            if hasattr(base, '__connection__') and not __connection__:
                __connection__ = base.__connection__
            if hasattr(base, '__tablename__') and not __tablename__:
                __tablename__ = base.__tablename__
            if hasattr(base, '__table__') and not __table__:
                __table__ = base.__table__
            if hasattr(base, '__tableclass__') and not __tableclass__:
                __tableclass__ = base.__tableclass__
            if hasattr(base, '__callbacks__') and not __callbacks__:
                __callbacks__ = base.__callbacks__

        # Add my own ORM attributes to pydantics base ModelMetaClass
        new_namespace = {
            '__connection__': __connection__,
            '__tablename__': __tablename__,
            '__table__': __table__,
            '__tableclass__': __tableclass__,
            '__callbacks__': __callbacks__,
            '__modelfields__': __modelfields__,
            #'__query__': {},
            #'_test1': 'hi',
            **{n: v for n, v in namespace.items()},
        }

        #dd(new_namespace)

        # Call pydantic ModelMetaClass
        # Amoung other things, pydantic will take all model attributes that do not
        # begin with a _ and convert them into ModelField classes
        # This is why I keep the originals in my new __modelfields__ attribute
        cls = super().__new__(mcls, name, bases, new_namespace, **kwargs)
        #dump(name, cls, cls.__dict__, '-----------------------------')

        # Meta is fired up more than once, sometimes pydantic has NOT
        # actually populated all fields.  If no fields, ignore rest of this custom __new__
        if not cls.__fields__: return cls

        # Pretty Printer
        # Register a pretty printer just for this entity.  Why not on the main
        # Model class itself?  Because then it prints each record as if it were
        # uvicore.orm.model.Model instead of the actual model class (ie: uvicore.auth.models.user.User)
        @register_pretty(cls)
        def pretty_entity(value, ctx):
            return pretty_call(ctx, cls, **value.__dict__)

        #dump("Registering Schema in Metaclass")

        # Build connection, tablename and table from tableclass
        if cls.__tableclass__ is not None:
            if cls.__connection__ is None: cls.__connection__ = cls.__tableclass__.connection
            if cls.__tablename__ is None: cls.__tablename__ = cls.__tableclass__.name
            if cls.__table__ is None: cls.__table__ = cls.__tableclass__.schema


        # Dynamically Build SQLAlchemy Table From Model Properties
        if cls.__table__ is not None:
            pass
            # Set connection and table name from table class
            #dd(cls.__table__.metadata.__dict__)
            #dd(dir(cls.__table__))
            #if cls.__connection__ is None: cls.__connection__ = cls.__table__.connection
            #if cls.__tablename__ is None: cls.__tablename__ = cls.__table__.name

        else:
            pass
            #dump('Building SA Table From Model Properties')
            # Fixme, code could go here to dynamically build
            # an SQLAlchemy table here.

            # An example of basic build, but fix to use __fields__
            # and each fields .field_info instead, and convert each model
            # property Field() into SA table stuff like columns and indexes...
            # columns = [x for x in entity.__field_defaults__.values()]
            # __class__.__table__ = sa.Table(
            #     entity.__tablename__,
            #     db.metadata.get(entity.__connection__),
            #     *columns
            # )

        # Pull out all callbacks from __modelfields__ and store in cls.__callbacks__ for future processing
        for (key, field) in cls.__modelfields__.items():
            if field.callback:
                callback = field.callback
                if type(callback) == str:
                    callback = getattr(cls, field.callback)
                cls.__callbacks__[key] = callback

        #dump(cls.__dict__)
        return cls


# IoC Class Instance
#ModelMetaclass: _ModelMetaclass = uvicore.ioc.make('ModelMetaclass', _ModelMetaclass)



from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, List, TypeVar

from uvicore.contracts import Application, Package

# Generic Router (APIRouter or WebRouter)
R = TypeVar('R')



# class Routes(ABC):

#     @abstractmethod
#     def register()







# Old direct routeint

# class Routes(Generic[R], ABC):

#     @property
#     @abstractmethod
#     def app(self) -> Application: pass

#     @property
#     @abstractmethod
#     def package(self) -> Package: pass

#     @property
#     @abstractmethod
#     def Router(self) -> R: pass

#     @property
#     @abstractmethod
#     def prefix(self) -> str: pass

#     @abstractmethod
#     def include(self, module, *, prefix: str = '', tags: List[str] = None) -> None:
#         """Include a new router object"""
#         pass



import uvicore
from base64 import b64decode
from uvicore.contracts import UserInfo
from uvicore.support.dumper import dump, dd
from uvicore.http.request import HTTPConnection
from uvicore.typing import Dict, Optional, Union
from uvicore.auth.authenticators.base import Authenticator


@uvicore.service()
class Basic(Authenticator):
    """HTTP basic auth authenticator"""

    # Notice:  Do not ever throw or return an error from authenticators.
    # If there is any auth problem (no headers, invalid or expired tokens, bad password)
    # then return True or False instead.  Any issues or invalid credentials causes the global
    # Authentication middleware to inject an Anonymous user, not to throw errors.
    # Any permissions errors happen at a route level when checking of the route
    # require an authenticated user or valid scope/role.

    # Return of False means this authentication method is not being attempted, try next authenticator
    # Return of True means this authentication method was being attempted, but failed validation, skip next authenticator
    # Return of User object means a valid user was found, skip next authenticator

    async def authenticate(self, request: HTTPConnection) -> Union[UserInfo, bool]:
        # 8500 req/sec
        #self.log.debug('BASIC Authenticator')
        # 7315 req/sec, logging is slow

        # Parse authorization header
        authorization, scheme, param = self.auth_header(request)

        # This authentication method not provided or attempted, goto next authenticator
        if not authorization or scheme != "basic":
            # Return of False means this authentication method is not being attempted
            # goto next authenticator in stack
            #self.log.debug('No Basic Auth header, goto next authenticator in stack')
            return False

        # Try to get the Basic Auth credentials
        try:
            data = b64decode(param).decode("ascii")
        except Exception:
            # No credentials defined from basic auth header, return True to denote Anonymous user and skip next authenticator
            self.log.debug('No credentials found in Basic Auth Header, seen as anonymous')
            return True

        # Get username and password from basic auth header
        username, separator, password = data.partition(":")

        # Incomplete username or password provided
        if not separator: return True

        # Get user and validate credentials
        user: UserInfo = await self.retrieve_user(username, password or '', self.config.provider, request)

        # Return user.  If no user return True to denote Anonymous User and skip next authenticator
        return user or True



import os
import uvicore
from uvicore.support import str
from uvicore.support.dumper import dump, dd
from uvicore.support.schematic import Schematic
from uvicore.console import command, argument, option


@command()
@argument('name')
async def table(name: str):
    """
    \b
    Generate a new Database table schematic...
    \b
    USAGE:
        Tables should be lower_underscore and PLURAL
        If a table is awkward as plural, ok to make a few singluar (user_info)
    \b
        ./uvicore gen table users
        ./uvicore gen table user_details
        ./uvicore gen table posts
        ./uvicore gen table post_tags
    """

    # Get calling package (main running app)
    package = uvicore.app.package(main=True)

    # Get stub (src)
    stub = os.path.dirname(__file__) + '/stubs/table.py'

    # Get destination for this filetype, considering the packages path customizations
    dest = package.folder_path('tables') + '/' + name + '.py'

    Schematic(
        type='table',
        stub=stub,
        dest=dest,
        replace = [
            ('xx_tablename', name),
            ('xx_TableName', str.studly(name)),
        ]
    ).generate()

    uvicore.log.nl()
    uvicore.log.notice('Be sure to add this table to your ./database/tables/__init__.py')


@command()
@argument('tablename')
@argument('modelname')
async def seeder(tablename: str, modelname: str):
    """
    \b
    Generate a new Database table seeder schematic...
    \b
    USAGE:
        Seeder should match the lower_underscore PLURAL tablename and lower_understore SINGLUAR model name
    \b
        ./uvicore gen seeder users user
        ./uvicore gen seeder user_details user_detail
        ./uvicore gen seeder posts post
        ./uvicore gen seeder post_tags post_tag
    """

    # Get calling package (main running app)
    package = uvicore.app.package(main=True)

    # Get stub (src)
    stub = os.path.dirname(__file__) + '/stubs/seeder.py'

    # Get destination for this filetype, considering the packages path customizations
    dest = package.folder_path('seeders') + '/' + tablename + '.py'

    Schematic(
        type='seeder',
        stub=stub,
        dest=dest,
        replace = [
            ('xx_tablename', tablename),
            ('xx_TableName', str.studly(tablename)),
            ('xx_modelname', modelname),
            ('xx_ModelName', str.studly(modelname)),
        ]
    ).generate()

    uvicore.log.nl()
    uvicore.log.notice('Be sure to add this seeder to your ./database/seeders/__init__.py')




import uvicore
import aiohttp
from uvicore.package import Provider
from uvicore.support.dumper import dump, dd


@uvicore.provider()
class HttpClient(Provider):

    def register(self) -> None:
        """Register package into the uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs, IoC bindings and early event listeners.
        Configs are deep merged only after all packages are registered.  No real
        work should be performed here as it is very early in the bootstraping
        process and we have no clear view of the full configuration system."""

        # Register event listeners
        # String based events instead of class based because HTTP may not even
        # be installed, so importing it would cause an issue.
        @uvicore.events.handle(['uvicore.console.events.command.Startup', 'uvicore.console.events.command.PytestStartup', 'uvicore.http.events.server.Startup'])
        async def uvicore_startup(event):
            #print('aiohttp console/http startup')
            # Fire up an aiohttp client session and bind to IoC as a singleton
            uvicore.ioc.bind('uvicore.http_client', aiohttp.ClientSession(),
                aliases=['http_client', 'aiohttp'],
                singleton=True
            )

        @uvicore.events.handle(['uvicore.console.events.command.Shutdown', 'uvicore.console.events.command.PytestShutdown', 'uvicore.http.events.server.Shutdown'])
        async def uvicore_shutdown(event):
            #print('aiohttp console/http shutdown')
            # Shutdown the aiohttp client session
            await uvicore.ioc.make('aiohttp').close()

    def boot(self) -> None:
        """Bootstrap package into the uvicore framework.
        Boot takes place after ALL packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all
        configuration. This is where you register, connections, models,
        views, assets, routes, commands...  If you need to perform work after ALL
        packages have booted, use the event system and listen to the booted event:
        self.events.listen('uvicore.foundation.events.app.Booted', self.booted)"""

        pass



import uvicore
from base64 import b64decode
from uvicore.http import status
from uvicore.typing import Tuple, Dict
from uvicore.http import Request
from uvicore.support import module
from uvicore.support.dumper import dump, dd
from fastapi.security import SecurityScopes
from uvicore.http.exceptions import HTTPException, PermissionDenied, NotAuthenticated, InvalidCredentials
from uvicore.auth.middleware.auth import Auth

@uvicore.service()
class Basic(Auth):
    """Basic HTTP Authentication Route Middleware"""

    def __init__(self, config: Dict):
        self.config = config

    async def __call__(self, scopes: SecurityScopes, request: Request):
        dump('BASIC HERE')
        dump(scopes.__dict__, self.config)

        # Assume unauthorized
        authorized = False

        # Parse authorization header
        authorization, scheme, param = self.auth_header(request)

        # Define Basic Auth unauthorized header only if we are allowed to return them, else None
        unauthorized_headers = None
        if self.config.return_www_authenticate_header:
            unauthorized_headers = {'WWW-Authenticate': 'Basic'}
            if self.config.realm: unauthorized_headers = {'WWW-Authenticate': 'Basic realm="{}"'.format(self.config.realm)}

        # This authorization method not provided or attempted, goto next guard in middleware stack
        if not authorization or scheme != "basic":
            if self.config.return_www_authenticate_header:
                # Only add WWW-Authenticate header if configured.  If Basic auth is the LAST of the guard stack
                # and we are in Web guard, we can safely prompt the browser Login box
                raise NotAuthenticated(headers=unauthorized_headers)
            else:
                # Return None means goto next middleware in guard stack
                return None

        # Try to get the Basic Auth credentials
        try:
            data = b64decode(param).decode("ascii")
        except Exception:
            # No credentials defined from basic auth header
            raise InvalidCredentials(headers=unauthorized_headers)
        username, separator, password = data.partition(":")

        # Incomplete username or password provided
        if not separator: raise InvalidCredentials(headers=unauthorized_headers)

        # Get user and validate credentials
        user = await self.retrieve_user(username, password or '', self.config.provider)

        # If no user returned, validation has failed or user not found
        if user is None: raise InvalidCredentials(headers=unauthorized_headers)

        # Hack logout by uncommenting this once
        #raise InvalidCredentials(headers=unauthorized_headers)

        # Validate Permissions
        self.validate_permissions(user, scopes)

        # Authorization successful.
        # Add user to request in case we use it in a decorator, we can pull it out with request.scope.get('user')
        request.scope['user'] = user

        # Return user in case we are using this guard as a dependency injected route parameter
        return user



from uvicore.typing import Dict


class Config(Dict):
    pass









# OBSOLETE, now SuperDict!


from abc import ABC, abstractmethod
from typing import Dict, Any
class ConfigOBSOLETE(ABC):

    @property
    @abstractmethod
    def items(self) -> Dict[str, Dict]:
        """Dictinoary of all packages deep_merged configs"""
        pass

    @abstractmethod
    def get(self, dotkey: str = None, _recursive_config: Dict = None) -> Any:
        """Get a config item by dotkey notation"""
        pass

    @abstractmethod
    def set(self, dotkey: str, value: any, _recursive_config: Dict = None) -> Any:
        """Set a config item by dotkey notation"""
        pass

    @abstractmethod
    def merge(self, dotkey: str, value: Any) -> None:
        """Deep merge value into all configs by dotkey notation"""
        pass



from __future__ import annotations

import operator as operators
from copy import deepcopy
from typing import Any, Dict, Generic, List, Tuple, TypeVar, Union, OrderedDict
from uvicore.support import hash

import sqlalchemy as sa
from sqlalchemy.sql.expression import BinaryExpression

from sqlalchemy.sql import quoted_name
from collections import OrderedDict as ODict
from dataclasses import dataclass

import uvicore
from uvicore.contracts import QueryBuilder as BuilderInterface
from uvicore.support.dumper import dd, dump

B = TypeVar("B")  # Builder Type (DbQueryBuilder or OrmQueryBuilder)
E = TypeVar("E")  # Entity Model


@uvicore.service()
class QueryBuilder(Generic[B, E], BuilderInterface[B, E]):

    def __init__(self):
        self.query = Query()

    def where(self, column: Union[str, BinaryExpression, List[Union[Tuple, BinaryExpression]]], operator: str = None, value: Any = None) -> B[B, E]:
        """Add where statement to query"""
        if type(column) == str or type(column) == sa.Column:
            # A single where as a string or actual SQLAlchemy Column
            # String default =
            #   .where('column', 'value')
            # String explicit operator:
            #   .where('column, '=', 'value')
            # SA Column default =
            #   .where(table.column, 'value')
            # SA Column explicit operator:
            #   .where(table.column, '=', 'value')

            # Swap operator and value
            if not value: value = operator; operator = '='
            self.query.wheres.append((column, operator.lower(), value))
        elif type(column) == list:
            # Multiple wheres in one as a List[Tuple] or List[BinaryExpression]
            for where in column:
                # Recursivelly add Tuple wheres
                if type(where) == tuple:
                    # String
                    #   .where([('column', 'value'), ('column', '=', 'value')])
                    # SA Column
                    #   .where([(table.column, 'value'), (table.column, '=', 'value)])
                    if len(where) == 2:
                        self.where(where[0], '=', where[1])
                    else:
                        self.where(where[0], where[1], where[2])
                else:
                    # SQLAlchemy Binary Expression
                    # .where([table.column == 'value', table.column >= 'value'])
                    self.where(where)
        else:
            # Direct SQLAlchemy expression
            # .where(table.column == 'value' and table.column >= 'value')
            self.query.wheres.append(column)
        return self

    def or_where(self, wheres: List[Union[Tuple, BinaryExpression]]) -> B[B, E]:
        """Add or where statement to query"""
        # Or where must be a list of tuple or BinaryExpression as it requires at least 2 statements
        # .or_where([('column', 'value'), ('column', '=', 'value')])
        # .or_where([table.column == value, table.column == value])
        or_wheres: List[Tuple] = []
        for where in wheres:
            if type(where) == tuple:
                if len(where) == 2:
                    or_wheres.append((where[0], '=', where[1]))
                else:
                    or_wheres.append((where[0], where[1].lower(), where[2]))
            else:
                # SQLAlchemy Binary Expression
                or_wheres.append(where)
        self.query.or_wheres.extend(or_wheres)
        return self

    def order_by(self, column: Union[str, List[str], List[Tuple], Any], order: str = 'ASC') -> B[B, E]:
        """Order results by these columns ASC or DESC order"""
        if type(column) == str:
            # A single order_by as a string or actual SQLAlchemy Column
            # .order_by('column') or .order_by('column', 'DESC')
            # .order_by
            self.query.order_by.append((column, order.upper()))
        elif type(column) == list:
            # Multiple order as a List[Tuple] (column, order)
            for order_by in column:
                if type(order_by) == tuple:
                    if len(order_by) == 1:
                        # Should never hit, because a tuple of (columnx) is actually not a tuple type at all
                        # but if someone added (columnx,) then it is a tuple, and order by should default to ASC
                        column = order_by[0]
                        order = 'ASC'
                    elif len(order_by) == 2:
                        column, order = order_by
                else:
                    column = order_by
                    order = 'ASC'
                self.order_by(column, order)
        else:
            # Direct SQLAlchemy expression
            if type(column) == sa.sql.elements.UnaryExpression:
                # If we use SA for order ASC or DESC then column is a UnaryExpression
                # .order_by(posts.id) or .order_by(sa.desc(posts.id)) or .order_by(sa.asc(posts.id))
                self.query.order_by.append(column)
            else:
                # If not UnaryExpression, then we assume column is a true SA colum, and we use our order string
                self.query.order_by.append((column, order))
        return self

    def limit(self, limit: int) -> B[B, E]:
        """Limit results"""
        self.query.limit = limit
        return self

    def offset(self, offset: int) -> B[B, E]:
        """Limit offset"""
        self.query.offset = offset
        return self

    def cache(self, key: str = None, *, seconds: int = None, store: str = None) -> B[B, E]:
        """Cache results, None seconds uses cache backend default, 0=forever"""
        # Seconds as None will default to cache configured default seconds
        self.query.cache = {
            'key': key,
            'seconds': seconds,
            'store': store,
        }
        return self

    def sql(self, method: str = 'select') -> str:
        """Get all SQL queries involved in this query builder"""
        query, saquery = self._build_query('select', self.query.copy())
        return str(saquery)

    def _build_query(self, method: str, query: Query) -> Tuple:
        # Convert our Query into SQLAlchemy query

        # where
        #   select
        #   update (not in insert)
        #   delete

        # insert will never come into this get() or build function
        if method == 'select':
            # Build .select() query from tables, joins and selectable columns
            saquery = self._build_select(query).distinct()

            # Build .select_from() query from tables and joins
            saquery = self._build_from(query, saquery)

            # Build .order_by() query
            saquery = self._build_order_by(query, saquery)

            # Build .group_by() queries
            saquery = self._build_group_by(query, saquery)

            # Build .limit() query
            if query.limit: saquery = saquery.limit(query.limit)

            # Build .offset query
            if query.offset: saquery = saquery.offset(query.offset)

        elif method == 'delete':
            # Build .delete() query from table
            saquery = sa.delete(query.table)

        elif method == 'update':
            # Build .update() query from table
            saquery = sa.update(query.table)

        # Build WHERE AND queries
        if query.wheres:
            where_ands = self._build_where(query, query.wheres)
            saquery = saquery.where(sa.and_(*where_ands))

        # Build WHERE OR queries
        if query.or_wheres:
            where_ors = self._build_where(query, query.or_wheres)
            saquery = saquery.where(sa.or_(*where_ors))


        # Return query and SQLAlchemy query
        return (query, saquery)

    def _build_group_by(self, query: Query, saquery):
        for column in query.group_by:
            column = self._column(column, query)
            saquery = saquery.group_by(column.sacol)
        return saquery

    def _build_order_by(self, query: Query, saquery):
        for order_by in query.order_by:
            if type(order_by) == tuple:
                column = self._column(order_by[0], query).sacol
                order = order_by[1]

                if order == 'DESC':
                    saquery = saquery.order_by(sa.desc(column))
                else:
                    saquery = saquery.order_by(column)
            else:
                # SQLAlchemy expression
                saquery = saquery.order_by(order_by)
        return saquery

    def _build_from(self, query: Query, saquery) -> sa.select.select_from:
        joins = query.table
        for join in query.joins:
            method = getattr(joins, join.method)
            joins = method(right=join.table, onclause=join.onclause)
        return saquery.select_from(joins)

    def _build_select(self, query: Query) -> sa.select:
        selects = []

        if not query.selects and not query.joins:
            # No explicit selects, no joins, return entire table of columns
            return sa.select([query.table])

        elif not query.selects and query.joins:
            selects.extend(query.table.columns)
            for join in query.joins:
                for column in join.table.columns:
                    selects.append(column.label(quoted_name(join.alias + '__' + column.name, True)))

        elif query.selects:
            # Explicit selects (can be on main table or joined relations)
            for select in query.selects:
                column = self._column(select, query)
                if column.alias:
                    selects.append(column.sacol.label(quoted_name(column.alias, True)))
                else:
                    selects.append(column.sacol)

        # Return SQLAlchemy .select() statment with above columns
        return sa.select(selects)

    def _build_where(self, query: Query, wheres: List[Tuple]):
        statements = []
        for where in wheres:
            if type(where) == tuple:
                column, operator, value = where
                column = self._column(column, query).sacol

                # Convert to SQL Alchemy Where
                if type(value) == str and value.lower() == 'null': value = None
                if operator == 'in':
                    statements.append(column.in_(value))
                elif operator == '!in':
                    statements.append(sa.not_(column.in_(value)))
                elif operator == 'like':
                    statements.append(column.like(value))
                elif operator == '!like':
                    statements.append(sa.not_(column.like(value)))
                else:
                    op = self._operator(operator)
                    statements.append(op(column, value))
            else:
                # SQLAlchemy expression
                statements.append(where)

        return statements

    def _connection(self):
        return self._conn

    def _pk(self):
        for column in self.query.table.primary_key.columns:
            # Just take first PK for now???
            return str(column.name)

    def _column(self, dotname: Any, query: Query = None) -> Column:
        # Column() class builder from dotname
        if not query: query = self.query
        if dotname is None: return None
        name = dotname
        alias = None
        table = query.table
        tablename = str(table.name)
        conn = self._connection()

        if type(dotname) == str:
            # Get column information from a string
            # This is separated into its own method so we can override it with the ORM builder
            table, tablename, column, name, conn = self._column_from_string(dotname, query)

        elif type(dotname) == sa.Column:
            # SQLAlchemy column
            column = dotname
            name = str(column.name)
            table = column.table
            tablename = str(column.table.name)
        else:
            # SQLAlchemy generic function (sa.func.count(), min(), max()...)
            column = dotname
            name = 'func'

        # Set alias only if using column is a relation
        if alias is None:
            if str(table.name) != str(query.table.name):
                alias = str(table.name) + '__' + name

        # Return new column class
        return Column(column, name, alias, conn, table, tablename)

    def _column_from_string(self, dotname: str, query: Query) -> Tuple:
        name = dotname
        table = query.table
        tablename = str(table.name)
        conn = self._connection()
        dotname = dotname.replace('__', '.')
        if '.' in dotname:
            parts = dotname.split('.')
            if len(parts) == 2:
                tablename, name = tuple(parts)
            elif len(parts) == 3:
                conn, tablename, name = tuple(parts)

            table = uvicore.db.table(tablename, conn)
            if table is None:
                # Table not found by name, must be a JOIN alias
                for join in query.joins:
                    if join.alias == tablename:
                        table = join.table
                        alias = join.alias + '__' + name
                        break;

        column = table.columns.get(name)
        return (table, tablename, column, name, conn)

    def _operator(self, operator: str):
        ops = {
            '=': operators.eq,
            '==': operators.eq,
            '!=': operators.ne,
            '>': operators.gt,
            '<': operators.lt,
            '>=': operators.ge,
            '<=': operators.le,
        }
        return ops[operator]




# IoC Class Instance
# No need to IoC this one because it is always inherited
# If you need to overrite it use the IoC to swap DbQueryBuilder or OrmQueryBuilder
# and set a new parent from there.


@dataclass
@uvicore.service()
class Column:
    # Private Column dataclass.  Not meant to be imported by any user but not _Column for display purposes
    # in logging and debugging.  I hate _ especially for dataclasses

    # __slots__ = (
    #     'sacol',
    #     'name',
    #     'alias',
    #     #'field',
    #     'connection',
    #     'table',
    #     'tablename',
    # )
    sacol: sa.Column
    name: str
    connection: str
    table: sa.Table
    tablename: str

    def __init__(self, sacol: sa.Column, name: str, alias: str, connection: str, table: sa.Table, tablename: str):
        self.sacol = sacol
        self.name = name
        self.alias = alias
        #self.field = field
        self.connection = connection
        self.table = table
        self.tablename = tablename


@dataclass
@uvicore.service()
class Join:
    # Private Join dataclass.  Not meant to be imported by any user but not _Join for display purposes
    # in logging and debugging.  I hate _ especially for dataclasses

    # __slots__ = (
    #     'table',
    #     'tablename',
    #     'left',
    #     'right',
    #     'onclause',
    #     'alias',
    #     'method'
    # )
    table: sa.Table
    tablename: str
    left: Column
    right: Column
    onclause: BinaryExpression
    alias: str
    method: str

    def __init__(self, table: sa.Table, tablename: str, left: Column, right: Column, onclause: BinaryExpression, alias: str, method: str):
        self.table = table
        self.tablename = tablename
        self.left = left
        self.right = right
        self.onclause = onclause
        self.alias = alias
        self.method = method


@dataclass
@uvicore.service()
class Query:
    # Private Query dataclass.  Not meant to be imported by any user but not _Query for display purposes
    # in logging and debugging.  I hate _ especially for dataclasses

    # __slots__ = (
    #     'includes',
    #     'selects',
    #     'wheres',
    #     'or_wheres',
    #     'filters',
    #     'or_filters',
    #     'group_by',
    #     'order_by',
    #     'sort',
    #     'limit',
    #     'offset',
    #     'keyed_by',
    #     'relations',
    #     'joins',
    #     'table',
    # )
    includes: List
    selects: List
    wheres: List
    or_wheres: List[Tuple]
    filters: List[Tuple]
    or_filters: List[Tuple]
    group_by: List
    order_by: List[Tuple]
    sort: List[Tuple]
    limit: Optional[int]
    offset: Optional[int]
    keyed_by: Optional[str]
    show_writeonly: Union[bool, List]
    cache: Dict
    relations: OrderedDict[str, Relation]
    joins: List[Join]
    table: sa.Table


    def __init__(self):
        self.includes: List = []
        self.selects: List = []
        self.wheres: List[Tuple] = []
        self.or_wheres: List[Tuple] = []
        self.filters: List[Tuple] = []
        self.or_filters: List[Tuple] = []
        self.group_by: List = []
        self.order_by: List[Tuple] = []
        self.sort: List[Tuple] = []
        self.limit: Optional[int] = None
        self.offset: Optional[int] = None
        self.keyed_by: Optional[str] = None
        self.show_writeonly: Union[bool, List] = False
        self.cache: Dict = None
        self.relations: OrderedDict[str, Relation] = ODict()
        self.joins: List[Join] = []
        self.table: sa.Table = None

    def copy(self):
        # Objects are always byref in python.  We want a complete deep clone
        # of a query.  Must use deep or it won't copy the .joins since its a
        # list of an actual class also.  Shallow copies all lists and dicts, but
        # not classes or list of classes, they will be byref unless deep
        table = self.table
        self.table = None
        newquery = deepcopy(self)
        self.table = table

        # Copy the original table by ref back or else SQLAlchemy will see a new
        # table class ID and think you are joining 2 different tables.  We must keep
        # the exact instance of each table.
        newquery.table = table
        return newquery

    def hash(self, *, hash_type: str = 'sha1', **kwargs) -> str:
        """Generate a unique hash for this query.  Used for automatic unique cache strings"""
        unique_params = {
            'tablename': self.table.name,
            'includes': self.includes,
            'selects': [str(col) for col in self.selects],
            'wheres': self.wheres,
            'or_wheres': self.or_wheres,
            'filters': self.filters,
            'or_filters': self.or_filters,
            'group_by': self.group_by,
            'order_by': self.order_by,
            'sort': self.sort,
            'limit': self.limit,
            'offset': self.offset,
            'keyed_by': self.keyed_by,
            'kwargs': kwargs,
        }
        hash_method = getattr(hash, hash_type)
        return hash_method(str(unique_params))




#from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Dict, Generic, TypeVar, Union, List, Tuple, Any

from uvicore.contracts import Mapper
from uvicore.contracts import OrmQueryBuilder
from uvicore.support.dumper import dump, dd

B = TypeVar("B")
E = TypeVar("E")

class Model(Generic[E], ABC):

    @abstractmethod
    def query(entity) -> OrmQueryBuilder[OrmQueryBuilder, E]:
        """ORM query builder passthrough"""

    # @abstractmethod
    # async def get(entity) -> Union[List[E], Dict[str, E]]:
    #     pass

    # @abstractmethod
    # async def find(entity, id: Any) -> Union[E, None]:
    #     pass

    # @abstractmethod
    # def include(entity, *args) -> OrmQueryBuilder[OrmQueryBuilder, E]:
    #     pass

    @abstractmethod
    async def insert(entity, models: Union[List[E], List[Dict]]) -> None:
        """Insert one or more entities as List of entities or List of Dictionaries

        This bulk insert does NOT allow inserting child relations at the
        same time as there is no way to get each parents PK out to
        reference with each child in BULK. If you want to insert parent
        and relations at the same time use the slower non-bulk
        insert_with_relations() instead.
        """

    @abstractmethod
    async def insert_with_relations(entity, models: List[Dict]) -> None:
        """Insert one or more entities as List of Dict that DO have relations included

        Because relations are included, this insert is NOT bulk and must
        loop each row, insert the parent, get the PK, then insert each
        children (or children first then parent depending on BelongsTo vs
        HasOne or HasMany)
        """

    @abstractmethod
    def mapper(self_or_entity, *args) -> Mapper:
        """Entity mapper for model->table or table->model conversions

        Can be accessed both from the [static] class or from an instance
        """

    @abstractmethod
    async def create(self, relation_name: str, models: Union[Any, List[Any]]) -> None:
        """Create related records and link them to this parent (self) model"""

    @abstractmethod
    async def add(self, relation_name: str, models: Union[Any, List[Any]]) -> None:
        """Alias to create"""

    @abstractmethod
    async def set(self, relation_name: str, models: Union[Any, List[Any]]) -> None:
        """Same as create(), except it deletes all first, so it sets the entire children"""

    @abstractmethod
    async def save(self) -> None:
        """Save this model to the database (insert or update)"""

    @abstractmethod
    async def delete(self) -> None:
        """Delete this model from the database"""

    @abstractmethod
    async def link(self, relation_name: str, models: Union[Any, List[Any]]) -> None:
        """Link records to relation using the Many-To-Many pivot table"""

    @abstractmethod
    async def unlink(self, relation_name: str, models: Union[Any, List[Any]] = None) -> None:
        """Unlink records to relation using the Many-To-Many pivot table"""

    @abstractmethod
    async def _before_insert(self) -> None:
        """Hook fired before record is inserted (new records only)"""

    @abstractmethod
    async def _after_insert(self) -> None:
        """Hook fired after record is inserted (new records only)"""

    @abstractmethod
    async def _before_save(self) -> None:
        """Hook fired before record is saved (inserted or updated)"""

    @abstractmethod
    async def _after_save(self) -> None:
        """Hook fired after record is saved (inserted or updated)"""

    @abstractmethod
    async def _before_delete(self) -> None:
        """Hook fired before record is deleted"""

    @abstractmethod
    async def _after_delete(self) -> None:
        """Hook fired after record is deleted"""
